import { Key } from 'aws-sdk/clients/dynamodb';
import { Context } from '../context';
import { DocumentWithId } from '../base/common';
import { CompositeCondition } from '../base/conditions';
/**
 * The query operator to use in the `KeyConditionExpression` to `QueryItem`
 * as called by [[find]]
 */
export declare type QueryOperator = 'match' | 'equals' | 'gte' | 'gt' | 'lte' | 'lt' | 'between';
/**
 * A find query. This is a map of key paths (specified dot-separated)
 * to values to match.
 *
 * The query you specify must match an access pattern on the collection,
 * otherwise an [[IndexNotFoundException]] will be thrown by [[find]]
 */
export declare type FindQuery = {
    [matchKey: string]: string | [string, string];
};
/**
 * The results of a [[find]] operation.
 */
export declare type FindResults<DocumentType extends DocumentWithId> = {
    /** The items in this batch of results */
    items: DocumentType[];
    /** The pagination token. This value is specified when
     * there are more results to fetch; pass it to another [[find]] call
     * to get the next batch. It will be `undefined` when there is no
     * more results.
     */
    nextToken?: Key;
};
/**
 * The options for a [[find]] operation
 */
export declare type FindOptions = {
    /**
     * The sort key condition to use
     */
    queryOperator?: QueryOperator;
    limit?: number;
    /**
     * `true` (default) to scan the index forward, `false` to scan it backward
     */
    scanForward?: boolean;
    /**
     * An optional filter expression for the
     * find operation
     */
    filter?: CompositeCondition;
};
/**
 * Find an item using one of its collection's access patterns
 *
 * This method is very strict about how the query object is used. Failure
 * to specify it
 *
 * Each of its keys is one of the keys paths to a field in the collection objects
 * (separated by a full-stop `.`), with the values set to the value to
 * match against. The value is a direct string comparison, so only string
 * values can be looked up (the query object cannot contain nested values
 * or values of types other than string)
 *
 * Every path that is specified must be part of the same access pattern.
 * You cannot have extraneous fields to match or mix keys from different
 * access patterns.
 *
 * All of the key paths in the partition key part of the access pattern
 * must be specified.
 *
 * You do not have to specify all of the sort keys,
 * **but** because they are stored as a composite key, and we can only do
 * `begins_with` match, you must specify the keys in the order they appear,
 * only leaving off those at the end.
 *
 * For example, if your sort key is defined as:
 *
 * `[['state'], ['city'], ['street', 'name'], ['street', 'number']]`
 *
 * you could specify any of the following combinations:
 * - `{ state: '...' }`
 * - `{ state: '...', city: '...' }`
 * - `{ state: '...', city: '...', 'street.name': '...' }`
 * - `{ state: '...', city: '...', 'street.name': '...', 'street.number': '...' }`
 *
 * But these would be invalid:
 * - `{ city: '...' }`
 * - `{ state: '...', 'street.name': '...' }`
 * - `{ city: '...', 'street.name': '...', 'street.number': '...' }`
 *
 *
 * @param ctx the context object
 * @param collectionName the collection name
 * @param query an object, with each of its keys set to the path of
 * each field in the search object to match and their corresponding
 * values to match (see above)s
 * @param nextToken pagination token
 * @param options options controlling the query
 * @param options.queryOperator query operator to use (defaults to `begins_with()`)
 * @param options.limit number of records to return
 * @param options.scanForward scan direction, true for forward across the index (default) or false for backward
 * @param options.filter a filter expression
 * @returns an object containing the items found and a pagination token
 * (if there is more results)
 * @throws [[`CollectionNotFoundException`]] when the collection is not found in the context
 * @throws [[`InvalidQueryException`]] when the access pattern cannot be found for the specfied combination of query key paths
 */
export declare function find<DocumentType extends DocumentWithId>(ctx: Context, collectionName: string, query: FindQuery, nextToken?: Key, options?: FindOptions): Promise<FindResults<DocumentType>>;
