import { Key } from 'aws-sdk/clients/dynamodb';
import { Context } from '../context';
import { DocumentWithId } from '../base/common';
import { CompositeCondition } from '../base/conditions';
/**
 * The results of a [[findChildren]] operation.
 */
export declare type FindChildrenResults<DocumentType extends DocumentWithId> = {
    /** the items that were returned in this batch */
    items: DocumentType[];
    /** The pagination token. If this value is specified, it means
     * there is more results for the query. Provide it to another
     * call to `findChildren` to get the next set of results.
     */
    nextToken?: Key;
};
export declare type KeyRangeExpression = {
    op: 'gte' | 'gt' | 'lte' | 'lt' | 'begins_with';
    value: string;
} | {
    op: 'between';
    min: string;
    max: string;
};
/**
 * The options to a [[findChildren]] operation
 */
export declare type FindChildrenOptions = {
    limit?: number;
    /**
     * `true` (default) to scan the index forward, `false` to scan it backward
     */
    scanForward?: boolean;
    /**
     * An optional filter expression for the
     * find operation
     */
    filter?: CompositeCondition;
    /**
     * The range of children to retrieve (on the child's ID value)
     */
    range?: KeyRangeExpression;
};
/**
 * Find all the child objects of a root (top-level) object.
 *
 * The parent collection is determined by the reference in the
 * child collection.
 *
 * This method has a `nextToken`, which is used for pagination - it
 * is returned when there is more values to retrieve. You should
 * write your code to repeatedly call findChildren with the nextToken
 * value of the previous call until it comes back undefined in order
 * to retrieve all the values.
 *
 * @param ctx the context
 * @param childCollectionName name of the child object collection
 * @param rootObjectId the `_id` of the root object
 * @param nextToken the next token from the previous call, or `undefined` if there are no more values
 * @param options the options to control the query
 * @param options.limit number of records to return
 * @param options.scanForward=true true for ascending index order
 * @param options.filter an optional filter expression to apply
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */
export declare function findChildren<DocumentType extends DocumentWithId>(ctx: Context, childCollectionName: string, rootObjectId: string, nextToken?: Key, options?: FindChildrenOptions): Promise<FindChildrenResults<DocumentType>>;
