import { Context } from '../context';
import { KeyPath } from '../base/access_pattern';
import { DocumentWithId } from '../base/common';
import { CompositeCondition } from '../base/conditions';
/**
 * An update object, where the key paths are specified as keys and the values
 * as the new field values.
 *
 * Keep in mind, that although this is a partial update, you need to specify
 * all the keys for a composite key in an access pattern - it cannot partially
 * update composite index values.
 */
export declare type SetValuesDocument = {
    [path: string]: any;
};
/**
 * A key path and value to set on a document as part
 * of an update operation. Specified as part of [[UpdateChangesDocument]],
 * this is a tuple in the form:
 *
 * `[key_path, new_value]`
 *
 * * *key_path* is the property path into the document to update (either
 *   as a dotted string path e.g. `'profile.name'` or a [[KeyPath]])
 * * *new_value* is the value to update the property path to (it cannot be
 *   undefined -- if you want to clear the value, see [[DeleteChange]])
 */
export declare type SetChange = [string | KeyPath, any];
/**
 * A property path to delete on a document as part of a [[UpdateChangesDocument]],
 * specified either as a dotted string path (e.g. `'profile.name'`) or a [[KeyPath]]
 */
export declare type DeleteChange = string | KeyPath;
/**
 * A set of changes to perform in an [[updateById]] or [[updateChildById]]
 * operation as an object. Each property is optional, but at
 * least one change must be specified.
 */
export declare type UpdateChangesDocument = {
    /**
     * The list of key paths to set a value. This is a list
     * of tuples, in the form [key_path, new_value]. The
     * key paths can be specified in a string form e.g. `'profile.name'`
     * or array form e.g. `['profile', 'name']`
     */
    $set?: SetChange[];
    /**
     * The list of key paths to clear the value. Key paths
     * may be specified in string form e.g. `'profile.name'` or
     * array form e.g. `['profile', 'name']`
     */
    $delete?: DeleteChange[];
};
/**
 * The set of updates to apply to a document. This can be
 * specified one of two ways:
 *
 * * an an object of key paths to values to set e.g. `{ 'profile.name': 'new name', 'status': 3 }`
 * * an operator object of changes to perform (see [[UpdateChangesDocument]])
 */
export declare type Updates = SetValuesDocument | UpdateChangesDocument;
/**
 * Update a document using its `_id`.
 *
 * This operation allows you to do a partial update of a collection document i.e. without
 * specifying all the values (it uses DynamoDB`s `UpdateItem` operation).
 *
 * At this time, the `updates` value just updates specified key paths on the target document.
 *
 * If some of the update key paths are indexed values, the indexes will also be updated. Because
 * of this, you must specify all the key values in an access pattern to ensure indexes are
 * updated consistently.
 *
 * @param ctx the context
 * @param collectionName the collection to update
 * @param objectId the `_id` value of the object to update
 * @param updates the set of updates to apply.
 * @returns the updated object value in its entirety.
 * @throws {CollectionNotFoundException} collection not found
 * @throws {InvalidUpdatesException} thrown when the updates object is invalid or incomplete
 * @throws {InvalidUpdateValueException} thrown when one of the update values is an invalid type
 */
export declare function updateById<DocumentType extends DocumentWithId>(ctx: Context, collectionName: string, objectId: string, updates: Updates, options?: {
    condition?: CompositeCondition;
}): Promise<DocumentType>;
