import { VError } from 'verror';
import isEqual from 'lodash/isEqual';
import { Converter } from 'aws-sdk/clients/dynamodb';
import get from 'lodash/get';
import isISOString from 'validator/lib/isISO8601';
import { randomBytes } from 'crypto';
import debug$1 from 'debug';
import fromPairs from 'lodash/fromPairs';

/**
 * Thrown when insert() is called with a specified _id
 * that already exists
 */
class ConflictException extends VError {
  constructor(message, id) {
    super({
      info: {
        id
      },
      name: 'conflict.error'
    }, message);
  }

}
/**
 * Thrown when an `_id` value is specified that is not
 * valid (_id values must be a string).
 */

class InvalidIdException extends VError {
  constructor(id) {
    super({
      info: {
        id
      },
      name: 'invalid_id.error'
    }, 'The provided document has an invalid ID');
  }

}
/**
 * Thrown when the child object to be inserted has a missing or invalid
 * parent key value.
 */

class InvalidParentIdException extends VError {
  constructor(parentId, collectionName, parentCollectionName) {
    super({
      info: {
        parentId,
        collectionName,
        parentCollectionName
      },
      name: 'invalid_parent_id.error'
    }, `The provided document has a missing parent ID or it is the incorrect type`);
  }

}
/**
 * Thrown when the collection specified cannot be found in the context, or
 * isn't of the expected type (root or child) for the called API.
 */

class CollectionNotFoundException extends VError {
  constructor(collection) {
    super({
      info: {
        collection
      },
      name: 'collection_not_found.error'
    }, `Collection not found: '${collection}'`);
  }

}
/**
 * Thrown when the index is not found for the provided query.
 */

class IndexNotFoundException extends VError {
  constructor(index) {
    super({
      info: {
        index
      },
      name: 'index_not_found.error'
    }, `Index not found: ${index}`);
  }

}
/**
 * Thrown during context construction to indicate a configuration
 * issue in the specified collections, access patterns or layouts.
 */

class ConfigurationException extends VError {
  constructor(message, options = {}) {
    super({ ...options,
      name: 'configuration.error'
    }, message);
  }

}
/**
 * Thrown on invalid field values that are provided during
 * persistence (insert/update/replace).
 */

class InvalidIndexedFieldValueException extends VError {
  constructor(message, {
    collection,
    keyPath
  }) {
    super({
      info: {
        collection,
        keyPath
      },
      name: 'invalid_indexed_field_value.error'
    }, message);
  }

}
/**
 * Thrown on an invalidly specified query provided to a
 * `find()` operation.
 */

class InvalidQueryException extends VError {
  constructor(message, {
    collection,
    query
  }) {
    super({
      info: {
        collection,
        query
      },
      name: 'invalid_query.error'
    }, message);
  }

}
/**
 * Thrown when the updates provided to an `update()` operation
 * are invalid.
 */

class InvalidUpdatesException extends VError {
  constructor(message) {
    super({
      name: 'invalid_updates.error'
    }, message);
  }

}
/**
 * Thrown when the update values provided in an `update()` operation
 * are invalid (e.g. `undefined`)
 */

class InvalidUpdateValueException extends VError {
  constructor(path, message) {
    super({
      name: 'invalid_update_value',
      info: {
        path
      }
    }, message);
  }

}
/**
 * Throw when the find descriptor for a transactFindByIds or
 * batchFindByIds is invalid.
 */

class InvalidFindDescriptorException extends VError {
  constructor(message) {
    super({
      name: 'invalid_find_descriptor'
    }, message);
  }

}
/**
 * Thrown when something from DynamoDB doesn't match
 * the context configuration during response processing
 */

class InternalProcessingException extends VError {
  constructor(message) {
    super({
      name: 'internal_processing'
    }, message);
  }

}
/**
 * @internal
 *
 * Print out a condition parser path for debugging / error handling
 * */

const printParsePath = parsePath => {
  return parsePath.map(e => {
    if (e.type === 'array') return e.index === 0 ? `[` : `[...@${e.index}:`;
    return `{ ${e.key}: `;
  }).join('');
};
/**
 * Thrown when there is a problem with the expression given
 * as a `FilterExpression` or `ConditionExpression`.
 */


class InvalidCompositeConditionException extends VError {
  constructor(message, parsePath) {
    super({
      name: 'invalid_composite_condition',
      info: {
        parsePath
      }
    }, `Condition parse exception: ${message} at ${printParsePath(parsePath)}`);
  }

}
/**
 * An exception thrown when a bad set of
 * [[BatchReplaceDeleteDescriptor]] is given to
 * [[batchReplaceDelete]]
 */

class InvalidBatchReplaceDeleteDescriptorException extends VError {
  constructor(message, info) {
    super({
      name: 'invalid_batch_replace_descriptor',
      info
    }, message);
  }

}
class InvalidRangeOperatorException extends VError {
  constructor(message, operator) {
    super({
      name: 'invalid_range_operator',
      info: {
        operator
      }
    }, message);
  }

}

/**
 * @internal
 *
 * Validate that the find keys specified for an index layout are valid.
 */

function validateFindKeys(findKeys) {
  const alreadyDefinedIndexes = [];
  findKeys.forEach((findKey, index) => {
    if (alreadyDefinedIndexes.includes(findKey.indexName)) {
      throw new ConfigurationException(`find key at index ${index} has duplicate index reference ${findKey.indexName}`);
    }

    alreadyDefinedIndexes.push(findKey.indexName);
  });
}

/**
 * A path to a key in a document. Each path element should be
 * a string. For example, a field called 'users' at the top level is
 * simply `['userId']`, while a nested field at 'description.title[0]'
 * is `['description', 'title', '0']`
 */

/** @internal */
const describeKeyPath = keyPath => keyPath.join('.');
/**
 * A function used to normalise a key value before it is packed into
 * an index value.
 *
 * This is useful when you want to index your fields differently to how
 * they're stored e.g. lowercase values before they are put in the
 * index.
 */

const describeAccessPattern = ({
  indexName,
  partitionKeys,
  sortKeys
}) => `[access pattern index=${indexName} partition=${partitionKeys.join(',')} ` + `${sortKeys ? `sort=${sortKeys.join(',')}` : ''}]`;

/**
 * @internal
 * Create the extract key value for setting up the context. An extract
 * key is an internal definition derived from access patterns to indicate
 * a key path that should be extracted for indexing upon insert.
 */
const withTypeCreateExtractKey = type => (key, valuePaths, options) => ({
  type,
  key,
  valuePaths,
  options: options || {}
});
/** @internal */

const createPartitionExtractKey = withTypeCreateExtractKey('partition');
/** @internal */

const createSortExtractKey = withTypeCreateExtractKey('sort');
/** @internal
 *
 * Construct the list of key paths to extract on `insert()` and `replace()` operations
 * for a collection, based on its access patterns.
 */

function buildAndValidateAccessPatterns(collection) {
  const wrapperExtractKeys = [];
  const findKeys = collection.layout.findKeys || [];
  const usedIndexes = [];

  for (const accessPattern of collection.accessPatterns || []) {
    const {
      indexName
    } = accessPattern;

    if (usedIndexes.includes(indexName)) {
      throw new ConfigurationException(`accessPattern ${describeAccessPattern(accessPattern)} refers to index in use by another pattern`);
    }

    usedIndexes.push(indexName);
    const layout = findKeys.find(key => key.indexName === indexName);

    if (!layout) {
      throw new ConfigurationException(`access pattern ${describeAccessPattern(accessPattern)} refers to index missing from layout`);
    }

    wrapperExtractKeys.push(createPartitionExtractKey(layout.partitionKey, accessPattern.partitionKeys, accessPattern.options));

    if (accessPattern.sortKeys) {
      if (!layout.sortKey) {
        throw new ConfigurationException(`access pattern ${describeAccessPattern(accessPattern)} has sort keys but index ${indexName} does not`);
      }

      wrapperExtractKeys.push(createSortExtractKey(layout.sortKey, accessPattern.sortKeys, accessPattern.options));
    } else if (!accessPattern.sortKeys && layout.sortKey) {
      throw new ConfigurationException(`access pattern ${describeAccessPattern(accessPattern)} does not ` + `have sort keys but index ${indexName} has one defined - values in this collection will not show up`);
    }
  }

  return wrapperExtractKeys;
}

/**
 * An opaquely defined type
 */

/**
 * Create a context object, with layouts and access patterns for
 * storing and retrieving data.
 *
 * @param dynamodb dynamodb instance, initialised with correct access key and region
 * @param collections a list of collection definitions to use with this context
 * @returns a context object
 *  @throws {ConfigurationException} when there is a configuration issue in the given collections
 */
function createContext(dynamodb, collections) {
  const definitions = new Map();
  const rootDefinitions = new Map();
  const childDefinitions = new Map();

  for (const collection of collections) {
    const {
      name,
      layout
    } = collection;

    if (definitions.has(name)) {
      throw new ConfigurationException(`Duplicate collection definition: '${name}'`);
    }

    if (layout.findKeys) {
      validateFindKeys(layout.findKeys);
    }

    let wrapperExtractKeys = [];

    if (collection.accessPatterns) {
      wrapperExtractKeys = buildAndValidateAccessPatterns(collection);
    }

    if (collection.ttlKeyPath) {
      const {
        ttlKeyPath
      } = collection;

      if (!layout.ttlAttribute) {
        throw new ConfigurationException(`Collection '${name}' defines ttlKeyPath=${describeKeyPath(ttlKeyPath)} but layout has no ttlAttribute specified`);
      }

      const ttlExtractKey = {
        type: 'ttl',
        key: layout.ttlAttribute,
        valuePaths: [ttlKeyPath],
        options: {}
      };
      wrapperExtractKeys = [...wrapperExtractKeys, ttlExtractKey];
    }

    definitions.set(collection.name, { ...collection,
      wrapperExtractKeys
    });

    if (collection.type === 'child') {
      childDefinitions.set(collection.name, { ...collection,
        wrapperExtractKeys
      });
    } else {
      rootDefinitions.set(collection.name, { ...collection,
        wrapperExtractKeys
      });
    }
  }

  for (const childDefinition of childDefinitions.values()) {
    const parentDefinition = rootDefinitions.get(childDefinition.parentCollectionName);

    if (!parentDefinition) {
      throw new ConfigurationException(`Child collection ${childDefinition.name} refers to non-existent parent definition ${childDefinition.parentCollectionName}`);
    }

    if (!isEqual(parentDefinition.layout, childDefinition.layout)) {
      throw new ConfigurationException(`Child collection ${childDefinition.name} must have same layout as parent definition ${parentDefinition.name}`);
    }
  }

  return {
    ddb: dynamodb,
    definitions,
    rootDefinitions,
    childDefinitions,
    __TYPE__: 'DynaglueContext'
  };
}

// Code adapted from https://github.com/mongodb/js-bson/blob/master/lib/objectid.js
/** @internal */

const PROCESS_UNIQUE = randomBytes(5);
/** @internal */

let IdIndex = ~~(Math.random() * 0xffffff);
/** @internal */

const getInc = () => IdIndex = (IdIndex + 1) % 0xffffff;
/**
 * @internal
 * Generate a new ID. Uses the Mongo BSON ID generation algorithm.
 */


const newId = () => {
  const time = ~~(Date.now() / 1000);
  const inc = getInc();
  const buffer = Buffer.alloc(12); // 4-byte timestamp

  buffer[3] = time & 0xff;
  buffer[2] = time >> 8 & 0xff;
  buffer[1] = time >> 16 & 0xff;
  buffer[0] = time >> 24 & 0xff; // 5-byte process unique

  buffer[4] = PROCESS_UNIQUE[0];
  buffer[5] = PROCESS_UNIQUE[1];
  buffer[6] = PROCESS_UNIQUE[2];
  buffer[7] = PROCESS_UNIQUE[3];
  buffer[8] = PROCESS_UNIQUE[4]; // 3-byte counter

  buffer[11] = inc & 0xff;
  buffer[10] = inc >> 8 & 0xff;
  buffer[9] = inc >> 16 & 0xff;
  return buffer.toString('hex');
};

/**
 * @internal
 * The separator value.
 */
const SEPARATOR = '|-|';
/**
 * @internal
 */

/**
 * @internal
 *
 * Assemble an _id value into its indexed primary key field
 */
const assemblePrimaryKeyValue = (collectionName, _id, separator = SEPARATOR) => `${collectionName}${separator}${_id}`; // FIXME: distinguish correctly between:
//  - sparse keys - a value is not always written to the all the key fields (from DynamoDB's persective, not optional sort
//                  key values) which leverages DynamoDB's behaviour to not index the record in that GSI
//  - empty key value - an access pattern mapped to a partition+sort defined GSI but that only has partition key values
//                   needs a "dummy" value in the sort key slot so it gets indexed, and not treated like a sort key

/**
 * @internal
 *
 * Assemble extracted key paths into their indexed value
 */

const assembleIndexedValue = (keyType, collectionName, values, separator = SEPARATOR) => {
  if (values.length === 0) {
    // empty key value
    return collectionName;
  } else if (values.every(value => typeof value === 'undefined')) {
    // sparse key? - keep the value blank to avoid showing up in searches
    return keyType === 'partition' ? collectionName : undefined;
  }

  return `${collectionName}${separator}${values.map(x => typeof x === 'string' ? x : '').join(separator)}`;
};
/**
 * @internal
 */

const getRootCollection = (context, collectionName) => {
  const c = context.rootDefinitions.get(collectionName);
  if (!c) throw new CollectionNotFoundException(collectionName);
  return c;
};
/**
 * @internal
 */

const getChildCollection = (context, collectionName) => {
  const c = context.childDefinitions.get(collectionName);
  if (!c) throw new CollectionNotFoundException(collectionName);
  return c;
};
/**
 * @internal
 */

const getCollection = (context, collectionName) => {
  const c = context.definitions.get(collectionName);
  if (!c) throw new CollectionNotFoundException(collectionName);
  return c;
};
/**
 * @internal
 *
 * Converts a TTL value from a document
 * to a DynamoDB compatible TTL value (UNIX date in seconds stored as a number)
 */

const transformTTLValue = ttlValue => {
  let transformedValue;

  if (typeof ttlValue === 'object' && ttlValue instanceof Date) {
    transformedValue = Math.ceil(ttlValue.getTime() / 1000);
  } else if (typeof ttlValue === 'number') {
    transformedValue = Math.ceil(ttlValue / 1000);
  } else if (typeof ttlValue === 'string' && isISOString(ttlValue)) {
    transformedValue = Math.ceil(new Date(ttlValue).getTime() / 1000);
  }

  return transformedValue;
};
/**
 * @internal
 *
 * Extract a transformed TTL value from a document
 */

const extractTransformedTTLValue = (value, valuePath) => {
  const ttlValue = get(value, valuePath);
  return transformTTLValue(ttlValue);
};
/**
 * @internal
 *
 * Given a collection, and set of key paths from an access pattern, create a value that can be used
 * to look up an index attribute. This is used to generate the value to store in the indexed attribute.
 */

const constructKeyValue = (type, collectionName, separator, valuePaths, options, value) => {
  if (type === 'ttl') {
    return extractTransformedTTLValue(value, valuePaths[0]);
  }

  const values = valuePaths.map(valuePath => {
    const extractedValue = get(value, valuePath);

    if (typeof extractedValue !== 'undefined' && typeof extractedValue !== 'string') {
      throw new InvalidIndexedFieldValueException(`Indexed value at path ${describeKeyPath(valuePath)} was not a string for collection ${collectionName}`, {
        keyPath: valuePath,
        collection: collectionName
      });
    }

    const transformedValue = options.stringNormalizer && extractedValue ? options.stringNormalizer(valuePath, extractedValue) : extractedValue;
    return transformedValue;
  });
  return assembleIndexedValue(type, collectionName, values, separator);
};
/**
 * @internal
 *
 * Generate the wrapped value of a document to store in a table for a collection.
 */

const toWrapped = (collection, value) => {
  let updatedValue;

  if (typeof value._id !== 'undefined') {
    if (typeof value._id !== 'string') {
      throw new InvalidIdException(value._id);
    }

    updatedValue = value;
  } else {
    const _id = collection.idGenerator ? collection.idGenerator() : newId();

    updatedValue = { ...value,
      _id
    };
  }

  const extractedKeys = collection.wrapperExtractKeys.map(({
    type,
    key,
    valuePaths,
    options
  }) => {
    var _collection$layout$in;

    const keyValue = constructKeyValue(type, collection.name, (_collection$layout$in = collection.layout.indexKeySeparator) !== null && _collection$layout$in !== void 0 ? _collection$layout$in : SEPARATOR, valuePaths, options, updatedValue);

    if (typeof keyValue !== undefined) {
      return {
        [key]: keyValue
      };
    }
  }).filter(x => typeof x !== 'undefined');
  let partitionKeyValue;
  let sortKeyValue;

  if (collection.type === 'child') {
    const parentId = get(value, collection.foreignKeyPath);

    if (typeof parentId !== 'string') {
      throw new InvalidParentIdException(parentId, collection.name, collection.parentCollectionName);
    }

    partitionKeyValue = assemblePrimaryKeyValue(collection.parentCollectionName, parentId, collection.layout.indexKeySeparator);
    sortKeyValue = assemblePrimaryKeyValue(collection.name, updatedValue._id, collection.layout.indexKeySeparator);
  } else {
    partitionKeyValue = sortKeyValue = assemblePrimaryKeyValue(collection.name, updatedValue._id, collection.layout.indexKeySeparator);
  }

  const wrapped = Object.assign({
    [collection.layout.primaryKey.partitionKey]: partitionKeyValue,
    [collection.layout.primaryKey.sortKey]: sortKeyValue,
    value: updatedValue,
    type: collection.name
  }, ...extractedKeys);
  return wrapped;
};
/**
 * @internal
 */

const unwrap = document => {
  return document.value;
};
/**
 * @internal
 */

const invertMap = map => Object.assign({}, ...Array.from(map.entries(), ([k, v]) => ({
  [v]: k
})));
/**
 * @internal
 */

const isSubsetOfKeyPath = (mainPath, subsetPath) => subsetPath.every((key, index) => mainPath[index] === key);
/**
 * @internal
 */

const findMatchingPath = (keyPaths, path) => {
  for (const keyPath of keyPaths) {
    if (isSubsetOfKeyPath(path, keyPath)) {
      return keyPath;
    }
  }
};

/** @internal */

const logger = debug$1('dynaglue:dynamodb');
/**
 * @internal
 *
 * Helper for logging dynamo requests
 */

function debugDynamo(operation, request) {
  logger('operation=%s request=%O', operation, request);
}

/**
 * Retrieve a top-level document by its `_id` field.
 *
 * @param context the context object
 * @param collectionName name of the collection to search
 * @param id the `_id` value of the root document
 * @returns the stored value, or `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */

async function findById(context, collectionName, id) {
  const collection = getRootCollection(context, collectionName);
  const request = {
    TableName: collection.layout.tableName,
    Key: Converter.marshall({
      [collection.layout.primaryKey.partitionKey]: assemblePrimaryKeyValue(collectionName, id, collection.layout.indexKeySeparator),
      [collection.layout.primaryKey.sortKey]: assemblePrimaryKeyValue(collectionName, id, collection.layout.indexKeySeparator)
    }, {
      convertEmptyValues: false
    })
  };
  debugDynamo('GetItem', request);
  const result = await context.ddb.getItem(request).promise();

  if (result.Item) {
    const wrapped = Converter.unmarshall(result.Item, {
      convertEmptyValues: false
    });
    return unwrap(wrapped);
  }

  return undefined;
}

/**
 * Insert a value into a collection. Adds an _id field to the value
 * if one is not provided using the `bson` ID generator (similar to
 * MongoDB IDs).
 *
 * @param context the context to use
 * @param collectionName name of the collection
 * @param value value to insert
 * @returns a copy of the inserted value, with appended `_id` field if not provided
 * @throws {ConflictException} when an item with the same _id already exists
 */

async function insert(context, collectionName, value) {
  const collection = getCollection(context, collectionName);
  const wrapped = toWrapped(collection, value);
  let request;

  if (collection.type === 'child') {
    request = {
      TableName: collection.layout.tableName,
      Item: Converter.marshall(wrapped, {
        convertEmptyValues: false
      }),
      ReturnValues: 'NONE',
      ConditionExpression: 'attribute_not_exists(#parentIdAttribute) and attribute_not_exists(#childIdAttribute)',
      ExpressionAttributeNames: {
        '#parentIdAttribute': assemblePrimaryKeyValue(collection.parentCollectionName, get(value, collection.foreignKeyPath), collection.layout.indexKeySeparator),
        '#childIdAttribute': assemblePrimaryKeyValue(collection.name, wrapped.value._id, collection.layout.indexKeySeparator)
      }
    };
  } else {
    request = {
      TableName: collection.layout.tableName,
      Item: Converter.marshall(wrapped, {
        convertEmptyValues: false
      }),
      ReturnValues: 'NONE',
      ConditionExpression: 'attribute_not_exists(#idAttribute)',
      ExpressionAttributeNames: {
        '#idAttribute': assemblePrimaryKeyValue(collection.name, wrapped.value._id, collection.layout.indexKeySeparator)
      }
    };
  }

  try {
    debugDynamo('PutItem', request);
    await context.ddb.putItem(request).promise();
  } catch (error) {
    if (error.code === 'ConditionalCheckFailedException') {
      throw new ConflictException('An item with this _id already exists', wrapped.value._id);
    }

    throw error;
  }

  return wrapped.value;
}

/* eslint-disable */
// prettier-ignore

/** @internal */
const DYNAMODB_RESERVED_WORDS = ['ABORT', 'ABSOLUTE', 'ACTION', 'ADD', 'AFTER', 'AGENT', 'AGGREGATE', 'ALL', 'ALLOCATE', 'ALTER', 'ANALYZE', 'AND', 'ANY', 'ARCHIVE', 'ARE', 'ARRAY', 'AS', 'ASC', 'ASCII', 'ASENSITIVE', 'ASSERTION', 'ASYMMETRIC', 'AT', 'ATOMIC', 'ATTACH', 'ATTRIBUTE', 'AUTH', 'AUTHORIZATION', 'AUTHORIZE', 'AUTO', 'AVG', 'BACK', 'BACKUP', 'BASE', 'BATCH', 'BEFORE', 'BEGIN', 'BETWEEN', 'BIGINT', 'BINARY', 'BIT', 'BLOB', 'BLOCK', 'BOOLEAN', 'BOTH', 'BREADTH', 'BUCKET', 'BULK', 'BY', 'BYTE', 'CALL', 'CALLED', 'CALLING', 'CAPACITY', 'CASCADE', 'CASCADED', 'CASE', 'CAST', 'CATALOG', 'CHAR', 'CHARACTER', 'CHECK', 'CLASS', 'CLOB', 'CLOSE', 'CLUSTER', 'CLUSTERED', 'CLUSTERING', 'CLUSTERS', 'COALESCE', 'COLLATE', 'COLLATION', 'COLLECTION', 'COLUMN', 'COLUMNS', 'COMBINE', 'COMMENT', 'COMMIT', 'COMPACT', 'COMPILE', 'COMPRESS', 'CONDITION', 'CONFLICT', 'CONNECT', 'CONNECTION', 'CONSISTENCY', 'CONSISTENT', 'CONSTRAINT', 'CONSTRAINTS', 'CONSTRUCTOR', 'CONSUMED', 'CONTINUE', 'CONVERT', 'COPY', 'CORRESPONDING', 'COUNT', 'COUNTER', 'CREATE', 'CROSS', 'CUBE', 'CURRENT', 'CURSOR', 'CYCLE', 'DATA', 'DATABASE', 'DATE', 'DATETIME', 'DAY', 'DEALLOCATE', 'DEC', 'DECIMAL', 'DECLARE', 'DEFAULT', 'DEFERRABLE', 'DEFERRED', 'DEFINE', 'DEFINED', 'DEFINITION', 'DELETE', 'DELIMITED', 'DEPTH', 'DEREF', 'DESC', 'DESCRIBE', 'DESCRIPTOR', 'DETACH', 'DETERMINISTIC', 'DIAGNOSTICS', 'DIRECTORIES', 'DISABLE', 'DISCONNECT', 'DISTINCT', 'DISTRIBUTE', 'DO', 'DOMAIN', 'DOUBLE', 'DROP', 'DUMP', 'DURATION', 'DYNAMIC', 'EACH', 'ELEMENT', 'ELSE', 'ELSEIF', 'EMPTY', 'ENABLE', 'END', 'EQUAL', 'EQUALS', 'ERROR', 'ESCAPE', 'ESCAPED', 'EVAL', 'EVALUATE', 'EXCEEDED', 'EXCEPT', 'EXCEPTION', 'EXCEPTIONS', 'EXCLUSIVE', 'EXEC', 'EXECUTE', 'EXISTS', 'EXIT', 'EXPLAIN', 'EXPLODE', 'EXPORT', 'EXPRESSION', 'EXTENDED', 'EXTERNAL', 'EXTRACT', 'FAIL', 'FALSE', 'FAMILY', 'FETCH', 'FIELDS', 'FILE', 'FILTER', 'FILTERING', 'FINAL', 'FINISH', 'FIRST', 'FIXED', 'FLATTERN', 'FLOAT', 'FOR', 'FORCE', 'FOREIGN', 'FORMAT', 'FORWARD', 'FOUND', 'FREE', 'FROM', 'FULL', 'FUNCTION', 'FUNCTIONS', 'GENERAL', 'GENERATE', 'GET', 'GLOB', 'GLOBAL', 'GO', 'GOTO', 'GRANT', 'GREATER', 'GROUP', 'GROUPING', 'HANDLER', 'HASH', 'HAVE', 'HAVING', 'HEAP', 'HIDDEN', 'HOLD', 'HOUR', 'IDENTIFIED', 'IDENTITY', 'IF', 'IGNORE', 'IMMEDIATE', 'IMPORT', 'IN', 'INCLUDING', 'INCLUSIVE', 'INCREMENT', 'INCREMENTAL', 'INDEX', 'INDEXED', 'INDEXES', 'INDICATOR', 'INFINITE', 'INITIALLY', 'INLINE', 'INNER', 'INNTER', 'INOUT', 'INPUT', 'INSENSITIVE', 'INSERT', 'INSTEAD', 'INT', 'INTEGER', 'INTERSECT', 'INTERVAL', 'INTO', 'INVALIDATE', 'IS', 'ISOLATION', 'ITEM', 'ITEMS', 'ITERATE', 'JOIN', 'KEY', 'KEYS', 'LAG', 'LANGUAGE', 'LARGE', 'LAST', 'LATERAL', 'LEAD', 'LEADING', 'LEAVE', 'LEFT', 'LENGTH', 'LESS', 'LEVEL', 'LIKE', 'LIMIT', 'LIMITED', 'LINES', 'LIST', 'LOAD', 'LOCAL', 'LOCALTIME', 'LOCALTIMESTAMP', 'LOCATION', 'LOCATOR', 'LOCK', 'LOCKS', 'LOG', 'LOGED', 'LONG', 'LOOP', 'LOWER', 'MAP', 'MATCH', 'MATERIALIZED', 'MAX', 'MAXLEN', 'MEMBER', 'MERGE', 'METHOD', 'METRICS', 'MIN', 'MINUS', 'MINUTE', 'MISSING', 'MOD', 'MODE', 'MODIFIES', 'MODIFY', 'MODULE', 'MONTH', 'MULTI', 'MULTISET', 'NAME', 'NAMES', 'NATIONAL', 'NATURAL', 'NCHAR', 'NCLOB', 'NEW', 'NEXT', 'NO', 'NONE', 'NOT', 'NULL', 'NULLIF', 'NUMBER', 'NUMERIC', 'OBJECT', 'OF', 'OFFLINE', 'OFFSET', 'OLD', 'ON', 'ONLINE', 'ONLY', 'OPAQUE', 'OPEN', 'OPERATOR', 'OPTION', 'OR', 'ORDER', 'ORDINALITY', 'OTHER', 'OTHERS', 'OUT', 'OUTER', 'OUTPUT', 'OVER', 'OVERLAPS', 'OVERRIDE', 'OWNER', 'PAD', 'PARALLEL', 'PARAMETER', 'PARAMETERS', 'PARTIAL', 'PARTITION', 'PARTITIONED', 'PARTITIONS', 'PATH', 'PERCENT', 'PERCENTILE', 'PERMISSION', 'PERMISSIONS', 'PIPE', 'PIPELINED', 'PLAN', 'POOL', 'POSITION', 'PRECISION', 'PREPARE', 'PRESERVE', 'PRIMARY', 'PRIOR', 'PRIVATE', 'PRIVILEGES', 'PROCEDURE', 'PROCESSED', 'PROJECT', 'PROJECTION', 'PROPERTY', 'PROVISIONING', 'PUBLIC', 'PUT', 'QUERY', 'QUIT', 'QUORUM', 'RAISE', 'RANDOM', 'RANGE', 'RANK', 'RAW', 'READ', 'READS', 'REAL', 'REBUILD', 'RECORD', 'RECURSIVE', 'REDUCE', 'REF', 'REFERENCE', 'REFERENCES', 'REFERENCING', 'REGEXP', 'REGION', 'REINDEX', 'RELATIVE', 'RELEASE', 'REMAINDER', 'RENAME', 'REPEAT', 'REPLACE', 'REQUEST', 'RESET', 'RESIGNAL', 'RESOURCE', 'RESPONSE', 'RESTORE', 'RESTRICT', 'RESULT', 'RETURN', 'RETURNING', 'RETURNS', 'REVERSE', 'REVOKE', 'RIGHT', 'ROLE', 'ROLES', 'ROLLBACK', 'ROLLUP', 'ROUTINE', 'ROW', 'ROWS', 'RULE', 'RULES', 'SAMPLE', 'SATISFIES', 'SAVE', 'SAVEPOINT', 'SCAN', 'SCHEMA', 'SCOPE', 'SCROLL', 'SEARCH', 'SECOND', 'SECTION', 'SEGMENT', 'SEGMENTS', 'SELECT', 'SELF', 'SEMI', 'SENSITIVE', 'SEPARATE', 'SEQUENCE', 'SERIALIZABLE', 'SESSION', 'SET', 'SETS', 'SHARD', 'SHARE', 'SHARED', 'SHORT', 'SHOW', 'SIGNAL', 'SIMILAR', 'SIZE', 'SKEWED', 'SMALLINT', 'SNAPSHOT', 'SOME', 'SOURCE', 'SPACE', 'SPACES', 'SPARSE', 'SPECIFIC', 'SPECIFICTYPE', 'SPLIT', 'SQL', 'SQLCODE', 'SQLERROR', 'SQLEXCEPTION', 'SQLSTATE', 'SQLWARNING', 'START', 'STATE', 'STATIC', 'STATUS', 'STORAGE', 'STORE', 'STORED', 'STREAM', 'STRING', 'STRUCT', 'STYLE', 'SUB', 'SUBMULTISET', 'SUBPARTITION', 'SUBSTRING', 'SUBTYPE', 'SUM', 'SUPER', 'SYMMETRIC', 'SYNONYM', 'SYSTEM', 'TABLE', 'TABLESAMPLE', 'TEMP', 'TEMPORARY', 'TERMINATED', 'TEXT', 'THAN', 'THEN', 'THROUGHPUT', 'TIME', 'TIMESTAMP', 'TIMEZONE', 'TINYINT', 'TO', 'TOKEN', 'TOTAL', 'TOUCH', 'TRAILING', 'TRANSACTION', 'TRANSFORM', 'TRANSLATE', 'TRANSLATION', 'TREAT', 'TRIGGER', 'TRIM', 'TRUE', 'TRUNCATE', 'TTL', 'TUPLE', 'TYPE', 'UNDER', 'UNDO', 'UNION', 'UNIQUE', 'UNIT', 'UNKNOWN', 'UNLOGGED', 'UNNEST', 'UNPROCESSED', 'UNSIGNED', 'UNTIL', 'UPDATE', 'UPPER', 'URL', 'USAGE', 'USE', 'USER', 'USERS', 'USING', 'UUID', 'VACUUM', 'VALUE', 'VALUED', 'VALUES', 'VARCHAR', 'VARIABLE', 'VARIANCE', 'VARINT', 'VARYING', 'VIEW', 'VIEWS', 'VIRTUAL', 'VOID', 'WAIT', 'WHEN', 'WHENEVER', 'WHERE', 'WHILE', 'WINDOW', 'WITH', 'WITHIN', 'WITHOUT', 'WORK', 'WRAPPED', 'WRITE', 'YEAR', 'ZONE'];
/* eslint-enable */

/**
 * @internal
 */

const isReservedWord = word => DYNAMODB_RESERVED_WORDS.includes(word.toUpperCase());
/**
 * @internal
 */

const isSafeAttributeName = attributeName => {
  if (isReservedWord(attributeName)) {
    return false;
  }

  return /^[A-Za-z][A-Za-z0-9]*$/.test(attributeName);
};

/**
 * @internal
 */

/**
 * @internal
 *
 * Create a mapper for generating `ExpressionAttributeNames`
 * entries. [[map]] will generate a new attribute name
 * that can be used in expressions for every attribute it
 * is given.
 *
 * The value for `ExpressionAttributeNames` can be
 * returned by [[get]] at the end.
 */
const createNameMapper = () => {
  let currentIndex = 0;
  const attributeNameMap = new Map();
  return {
    /**
     * Generate an expression attribute name for
     * `name` (if necessary - values not requiring
     * escaping will be returned as-is)
     */
    map(name, mappedName) {
      if (!mappedName && isSafeAttributeName(name)) {
        return name;
      }

      let nameMapping = attributeNameMap.get(name);

      if (!nameMapping) {
        nameMapping = mappedName !== null && mappedName !== void 0 ? mappedName : `#attr${currentIndex++}`;
        attributeNameMap.set(name, nameMapping);
      }

      return nameMapping;
    },

    /**
     * Return the map of attribute names
     */
    get() {
      const result = invertMap(attributeNameMap);
      if (Object.keys(result).length === 0) return undefined;
      return result;
    }

  };
};
/**
 * @internal
 */

/**
 * @internal
 *
 * Create a mapper for generating `ExpressionAttributeValues`
 * entries. [[map]] will generate a new attribute name
 * that can be used in expressions for every attribute it
 * is given.
 *
 * The value for `ExpressionAttributeValues` can be
 * returned by [[get]] at the end.
 */
const createValueMapper = () => {
  let currentIndex = 0;
  const valueMap = new Map();
  return {
    /**
     * Given `value`, marshall it to DynamoDB format, store
     * it internally, and return the `:value` reference that
     * can be used in expressions
     */
    map(value) {
      const valueKey = `:value${currentIndex++}`;
      const convertedValue = Converter.input(value, {
        convertEmptyValues: false
      });
      valueMap.set(valueKey, convertedValue);
      return valueKey;
    },

    /**
     * Get the map for `ExpressionAttributeValues`
     */
    get() {
      if (valueMap.size === 0) return undefined;
      return Array.from(valueMap).reduce((obj, [key, value]) => {
        obj[key] = value;
        return obj;
      }, {});
    }

  };
};

/**
 * @internal
 * Is one of the keys in the object a condition operator
 */
const isConditionKey = key => ['$or', '$and', '$eq', '$neq', '$gt', '$gte', '$lt', '$lte', '$between', '$in', '$exists', '$type', '$beginsWith', '$contains'].includes(key);
/**
 * @internal
 * Convert a string to a key path, adding the `value` object prefix.
 */


const mapKeyPath = (key, nameMapper) => [nameMapper.map('value', '#value'), ...key.split('.').map(pathElement => nameMapper.map(pathElement))].join('.');
/**
 * @internal
 *
 * Parse an object expression and give back a filter/condition expression
 * for DynamoDB.
 *
 * @param clause the expression object
 * @param context a context - pass your current NameMapper and ValueMapper, which will be filled out with filter expression parts
 * @returns the condition/filter expression
 */


const parseCompositeCondition = (clause, context) => {
  /* eslint-disable @typescript-eslint/no-use-before-define */
  if (Object.keys(clause).length === 1) {
    if ('$and' in clause) {
      return parseAndCondition(clause, context);
    } else if ('$or' in clause) {
      return parseOrCondition(clause, context);
    } else if ('$not' in clause) {
      return parseNotCondition(clause, context);
    }
  }

  return parseKeyPathsAndClause(clause, context);
  /* eslint-enable @typescript-eslint/no-use-before-define */
};
/**
 * @internal
 * Parse an AND condition expression
 */

const parseAndCondition = (clause, context) => {
  const parsePath = [...context.parsePath, {
    type: 'object',
    key: '$and'
  }];

  const updatedContextForIndex = index => ({ ...context,
    parsePath: [...parsePath, {
      type: 'array',
      index
    }]
  });

  const subclauses = clause.$and.map((clause, index) => parseCompositeCondition(clause, updatedContextForIndex(index)));
  return subclauses.map(clause => `(${clause})`).join(' AND ');
};
/**
 * @internal
 * Parse an OR condition expression
 */


const parseOrCondition = (clause, context) => {
  const parsePath = [...context.parsePath, {
    type: 'object',
    key: '$or'
  }];

  const updatedContextForIndex = index => ({ ...context,
    parsePath: [...parsePath, {
      type: 'array',
      index
    }]
  });

  const subclauses = clause.$or.map((clause, index) => parseCompositeCondition(clause, updatedContextForIndex(index)));
  return subclauses.map(clause => `(${clause})`).join(' OR ');
};
/**
 * @internal
 * Parse an NOT condition expression.
 */


const parseNotCondition = (clause, context) => {
  const parsePath = [...context.parsePath, {
    type: 'object',
    key: '$not'
  }];
  const updatedContext = { ...context,
    parsePath
  };
  const subclause = parseCompositeCondition(clause.$not, updatedContext);
  return `NOT ${subclause}`;
};
/**
 * @internal
 *
 * Parse a key paths object (one that has key paths as keys and
 * operators as values.
 */


const parseKeyPathsAndClause = (clause, context) => {
  const {
    nameMapper,
    valueMapper,
    parsePath
  } = context;
  const paths = Object.keys(clause);

  if (paths.length < 1) {
    throw new InvalidCompositeConditionException('expected at least one key path with operator', parsePath);
  }

  const conditionKey = paths.find(isConditionKey);

  if (conditionKey) {
    const keyParsePath = [...parsePath, {
      type: 'object',
      key: conditionKey
    }];
    throw new InvalidCompositeConditionException(`unexpected condition key`, keyParsePath);
  }

  const clauses = [];
  Object.entries(clause).forEach(([path, condition]) => {
    const keyParsePath = [...parsePath, {
      type: 'object',
      key: path
    }];
    let clauseString;

    const simpleClause = (operator, value) => {
      const valueName = valueMapper.map(value);
      return `${mapKeyPath(path, nameMapper)} ${operator} ${valueName}`;
    };

    if ('$eq' in condition) {
      const value = condition.$eq;
      clauseString = simpleClause('=', value);
    } else if ('$neq' in condition) {
      const value = condition.$neq;
      clauseString = simpleClause('<>', value);
    } else if ('$gt' in condition) {
      const value = condition.$gt;
      clauseString = simpleClause('>', value);
    } else if ('$gte' in condition) {
      const value = condition.$gte;
      clauseString = simpleClause('>=', value);
    } else if ('$lt' in condition) {
      const value = condition.$lt;
      clauseString = simpleClause('<', value);
    } else if ('$lte' in condition) {
      const value = condition.$lte;
      clauseString = simpleClause('<=', value);
    } else if ('$between' in condition) {
      const value = condition.$between;

      if (!value || typeof value !== 'object' || typeof value.$lte === 'undefined' || typeof value.$gte === 'undefined') {
        throw new InvalidCompositeConditionException('$between must be an object with values for $lte and $gte', keyParsePath);
      }

      const {
        $lte,
        $gte
      } = condition.$between;
      const value1 = valueMapper.map($gte);
      const value2 = valueMapper.map($lte);
      clauseString = `${mapKeyPath(path, nameMapper)} BETWEEN ${value1} AND ${value2}`;
    } else if ('$in' in condition) {
      if (!Array.isArray(condition.$in)) {
        throw new InvalidCompositeConditionException('$in must be an array of values', keyParsePath);
      }

      const values = condition.$in;

      if (values.length > 100) {
        throw new InvalidCompositeConditionException('$in condition has too many values', keyParsePath);
      } else if (values.length === 0) {
        throw new InvalidCompositeConditionException('$in condition must have at least one value', keyParsePath);
      }

      const valueNames = values.map(value => valueMapper.map(value));
      clauseString = `${mapKeyPath(path, nameMapper)} IN (${valueNames.join(',')})`;
    } else if ('$exists' in condition) {
      const fn = condition.$exists ? 'attribute_exists' : 'attribute_not_exists';
      clauseString = `${fn}(${mapKeyPath(path, nameMapper)})`;
    } else if ('$type' in condition) {
      const valueName = valueMapper.map(condition.$type);
      clauseString = `attribute_type(${mapKeyPath(path, nameMapper)},${valueName})`;
    } else if ('$beginsWith' in condition) {
      const valueName = valueMapper.map(condition.$beginsWith);
      clauseString = `begins_with(${mapKeyPath(path, nameMapper)},${valueName})`;
    } else if ('$contains' in condition) {
      const valueName = valueMapper.map(condition.$contains);
      clauseString = `contains(${mapKeyPath(path, nameMapper)},${valueName})`;
    } else throw new InvalidCompositeConditionException('unknown operator', keyParsePath);

    clauses.push(clauseString);
  });
  return clauses.join(' AND ');
};

/**
 * The query operator to use in the `KeyConditionExpression` to `QueryItem`
 * as called by [[find]]
 */

/**
 * @internal
 *
 * Compare key paths and return if they are equal.
 */
const isEqualKey = (lhs, rhs) => {
  const normalisedLhs = typeof lhs === 'string' ? lhs.split('.') : lhs;
  const normalisedRhs = typeof rhs === 'string' ? rhs.split('.') : rhs;
  return isEqual(normalisedLhs, normalisedRhs);
};
/**
 * @internal
 *
 * Find the access pattern for the specified query.
 */

const findAccessPattern = (collection, query) => {
  return (collection.accessPatterns || []).find(ap => {
    const unmatchedQueryKeys = [...Object.keys(query)];

    for (const apKey of ap.partitionKeys) {
      const matchingQueryKeyIndex = unmatchedQueryKeys.findIndex(queryKey => isEqualKey(apKey, queryKey));

      if (matchingQueryKeyIndex >= 0) {
        unmatchedQueryKeys.splice(matchingQueryKeyIndex, 1);
      } else {
        return false;
      }
    }

    if (ap.sortKeys) {
      for (const apKey of ap.sortKeys) {
        const matchingQueryKeyIndex = unmatchedQueryKeys.findIndex(queryKey => isEqualKey(apKey, queryKey));

        if (matchingQueryKeyIndex >= 0) {
          unmatchedQueryKeys.splice(matchingQueryKeyIndex, 1);
        } else {
          break;
        }
      }
    }

    return unmatchedQueryKeys.length === 0;
  });
};
/**
 * @internal
 *
 * Find the access pattern layout for the specified access pattern
 */

const findAccessPatternLayout = (findKeys, ap) => findKeys.find(fk => fk.indexName === ap.indexName);
/**
 * @internal
 *
 * Assemble a value into the query, taking into account string normalizer options.
 */

const assembleQueryValue = (type, collectionName, query, options, paths, separator) => {
  if (paths) {
    const values = [];

    for (const path of paths) {
      const keyPath = path.join('.');
      const pathValue = query[keyPath];
      if (!pathValue) break;
      const scalarPathValue = type === 'partition' || type === 'sort' ? Array.isArray(pathValue) ? pathValue[0] : pathValue : Array.isArray(pathValue) ? pathValue[1] : undefined;
      if (!scalarPathValue) throw new InvalidIndexedFieldValueException('$between queries must specify an array for sort key values', {
        collection: collectionName,
        keyPath: path
      });
      const transformedValue = options.stringNormalizer ? options.stringNormalizer(path, scalarPathValue) : scalarPathValue;
      values.push(transformedValue);
    }

    return assembleIndexedValue(type === 'sort2' ? 'sort' : type, collectionName, values, separator);
  }

  return undefined;
}; // FIXME: distinguish `=` and `begins_with` based on specified sort keys

/**
 * @internal
 */

const getQueryOperator = (sortKeyName, sortValueName, sortValueName2, qo) => {
  switch (qo) {
    case 'lte':
      return `${sortKeyName} <= ${sortValueName}`;

    case 'lt':
      return `${sortKeyName} < ${sortValueName}`;

    case 'gte':
      return `${sortKeyName} >= ${sortValueName}`;

    case 'gt':
      return `${sortKeyName} > ${sortValueName}`;

    case 'equals':
      return `${sortKeyName} = ${sortValueName}`;

    case 'between':
      return `${sortKeyName} BETWEEN ${sortValueName} AND ${sortValueName2}`;

    default:
    case 'match':
      return `begins_with(${sortKeyName}, ${sortValueName})`;
  }

  throw new Error('unreachable');
};
/**
 * The options for a [[find]] operation
 */


/**
 * Find an item using one of its collection's access patterns
 *
 * This method is very strict about how the query object is used. Failure
 * to specify it
 *
 * Each of its keys is one of the keys paths to a field in the collection objects
 * (separated by a full-stop `.`), with the values set to the value to
 * match against. The value is a direct string comparison, so only string
 * values can be looked up (the query object cannot contain nested values
 * or values of types other than string)
 *
 * Every path that is specified must be part of the same access pattern.
 * You cannot have extraneous fields to match or mix keys from different
 * access patterns.
 *
 * All of the key paths in the partition key part of the access pattern
 * must be specified.
 *
 * You do not have to specify all of the sort keys,
 * **but** because they are stored as a composite key, and we can only do
 * `begins_with` match, you must specify the keys in the order they appear,
 * only leaving off those at the end.
 *
 * For example, if your sort key is defined as:
 *
 * `[['state'], ['city'], ['street', 'name'], ['street', 'number']]`
 *
 * you could specify any of the following combinations:
 * - `{ state: '...' }`
 * - `{ state: '...', city: '...' }`
 * - `{ state: '...', city: '...', 'street.name': '...' }`
 * - `{ state: '...', city: '...', 'street.name': '...', 'street.number': '...' }`
 *
 * But these would be invalid:
 * - `{ city: '...' }`
 * - `{ state: '...', 'street.name': '...' }`
 * - `{ city: '...', 'street.name': '...', 'street.number': '...' }`
 *
 *
 * @param ctx the context object
 * @param collectionName the collection name
 * @param query an object, with each of its keys set to the path of
 * each field in the search object to match and their corresponding
 * values to match (see above)s
 * @param nextToken pagination token
 * @param options options controlling the query
 * @param options.queryOperator query operator to use (defaults to `begins_with()`)
 * @param options.limit number of records to return
 * @param options.scanForward scan direction, true for forward across the index (default) or false for backward
 * @param options.filter a filter expression
 * @returns an object containing the items found and a pagination token
 * (if there is more results)
 * @throws [[`CollectionNotFoundException`]] when the collection is not found in the context
 * @throws [[`InvalidQueryException`]] when the access pattern cannot be found for the specfied combination of query key paths
 */
async function find(ctx, collectionName, query, nextToken, options = {}) {
  var _collection$layout$fi, _collection$layout, _options$scanForward;

  const collection = getCollection(ctx, collectionName);
  const ap = findAccessPattern(collection, query);

  if (!ap) {
    throw new InvalidQueryException('Unable to find access pattern matching query', {
      collection: collectionName,
      query
    });
  }

  const layout = findAccessPatternLayout((_collection$layout$fi = (_collection$layout = collection.layout) === null || _collection$layout === void 0 ? void 0 : _collection$layout.findKeys) !== null && _collection$layout$fi !== void 0 ? _collection$layout$fi : [], ap);

  if (!layout) {
    throw new ConfigurationException(`Unable to find layout for index specified in access pattern`, {
      info: {
        collectionName,
        indexName: ap.indexName
      }
    });
  }

  const partitionKeyValue = assembleQueryValue('partition', collection.name, query, ap.options || {}, ap.partitionKeys, collection.layout.indexKeySeparator);
  const sortKeyValue = assembleQueryValue('sort', collection.name, query, ap.options || {}, ap.sortKeys, collection.layout.indexKeySeparator);
  const sortKeyValue2 = (options === null || options === void 0 ? void 0 : options.queryOperator) === 'between' ? assembleQueryValue('sort2', collection.name, query, ap.options || {}, ap.sortKeys, collection.layout.indexKeySeparator) : undefined;
  const nameMapper = createNameMapper();
  const valueMapper = createValueMapper();
  const sortKeyOp = sortKeyValue && getQueryOperator( // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
  nameMapper.map(layout.sortKey, '#indexSortKey'), valueMapper.map(sortKeyValue), (options === null || options === void 0 ? void 0 : options.queryOperator) === 'between' ? valueMapper.map(sortKeyValue2) : undefined, options === null || options === void 0 ? void 0 : options.queryOperator);
  const keyConditionExpression = `${nameMapper.map(layout.partitionKey, '#indexPartitionKey')} = ${valueMapper.map(partitionKeyValue)}${sortKeyValue ? ` AND ${sortKeyOp}` : ''}`;
  let filterExpression;

  if (options !== null && options !== void 0 && options.filter) {
    filterExpression = parseCompositeCondition(options.filter, {
      nameMapper,
      valueMapper,
      parsePath: []
    });
  }

  const queryRequest = {
    TableName: collection.layout.tableName,
    IndexName: ap.indexName,
    KeyConditionExpression: keyConditionExpression,
    ExpressionAttributeNames: nameMapper.get(),
    ExpressionAttributeValues: valueMapper.get(),
    ExclusiveStartKey: nextToken,
    Limit: options === null || options === void 0 ? void 0 : options.limit,
    ScanIndexForward: (_options$scanForward = options === null || options === void 0 ? void 0 : options.scanForward) !== null && _options$scanForward !== void 0 ? _options$scanForward : true,
    FilterExpression: filterExpression
  };
  debugDynamo('Query', queryRequest);
  const {
    Items: items,
    LastEvaluatedKey: lastEvaluatedKey
  } = await ctx.ddb.query(queryRequest).promise();
  const unwrappedItems = items ? items.map(item => unwrap(Converter.unmarshall(item, {
    convertEmptyValues: false
  }))) : [];
  return {
    items: unwrappedItems,
    nextToken: lastEvaluatedKey
  };
}

/**
 * Delete a root object using its `_id` field
 * @param context the context object
 * @param collectionName the name of the collection
 * @param id the object to remove
 * @param options options to apply
 * @param options.condition an optional conditional expression that must be satifisfied for the update to proceed
 * @returns the deleted object (as stored in the database), or
 * `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */

async function deleteById(context, collectionName, id, options = {}) {
  const collection = getRootCollection(context, collectionName);
  const nameMapper = createNameMapper();
  const valueMapper = createValueMapper();
  let conditionExpression;

  if (options.condition) {
    conditionExpression = parseCompositeCondition(options.condition, {
      nameMapper,
      valueMapper,
      parsePath: []
    });
  }

  const request = {
    TableName: collection.layout.tableName,
    Key: Converter.marshall({
      [collection.layout.primaryKey.partitionKey]: assemblePrimaryKeyValue(collectionName, id, collection.layout.indexKeySeparator),
      [collection.layout.primaryKey.sortKey]: assemblePrimaryKeyValue(collectionName, id, collection.layout.indexKeySeparator)
    }, {
      convertEmptyValues: false
    }),
    ReturnValues: 'ALL_OLD',
    ConditionExpression: conditionExpression,
    ExpressionAttributeNames: nameMapper.get(),
    ExpressionAttributeValues: valueMapper.get()
  };
  debugDynamo('DeleteItem', request);
  const result = await context.ddb.deleteItem(request).promise();

  if (result.Attributes) {
    const wrapped = Converter.unmarshall(result.Attributes, {
      convertEmptyValues: false
    });
    return unwrap(wrapped);
  }

  return undefined;
}

/**
 * Insert or replace a value in a collection.
 *
 * This operation differs from [[insert]] in that it does not check for the existence
 * of a document with the same `_id` value - it will replace whatever is there.
 *
 * @param context the context
 * @param collectionName the collection to update
 * @param value the document to insert or replace
 * @param options options to apply
 * @param options.condition an optional conditional expression that must be satifisfied for the update to proceed
 * @returns the inserted / replaced value
 * @throws {CollectionNotFoundException} when the collection is not found
 */

async function replace(context, collectionName, value, options = {}) {
  const collection = getCollection(context, collectionName);
  const wrapped = toWrapped(collection, value);
  let conditionExpression;
  const nameMapper = createNameMapper();
  const valueMapper = createValueMapper();

  if (options.condition) {
    conditionExpression = parseCompositeCondition(options.condition, {
      nameMapper,
      valueMapper,
      parsePath: []
    });
  }

  const request = {
    TableName: collection.layout.tableName,
    Item: Converter.marshall(wrapped),
    ReturnValues: 'NONE',
    ConditionExpression: conditionExpression,
    ExpressionAttributeNames: nameMapper.get(),
    ExpressionAttributeValues: valueMapper.get()
  };
  debugDynamo('PutItem', request);
  await context.ddb.putItem(request).promise();
  return wrapped.value;
}

/**
 * @private
 * Increment the last character to one character lexigraphically higher. It
 * might not be a valid Unicode character, but that shouldn't matter as we're
 * only using them for range expressions.
 * 
 * NOTE: I'm not sure if this is 100% correct as Unicode codepoints can go
 * up to 0x11000, but working on the boundary condition of 0x10FFF isn't
 * a valid codepoint so JavaScript falls back to code-unit behaviour. For
 * DynamoDB purposes, this may need to be done as a straight Buffer man-
 * ipulation.
 *
 * @param x  the string to increment the last character of
 * @returns the transformed string
 */
const incrementLast = x => {
  var _x$codePointAt;

  if (x.length === 0) {
    return '\u0000';
  }

  const prefix = x.slice(0, x.length - 1);
  const lastCodePoint = (_x$codePointAt = x[x.length - 1].codePointAt(0)) !== null && _x$codePointAt !== void 0 ? _x$codePointAt : 0;

  if (lastCodePoint >= 0xffff) {
    return incrementLast(prefix) + String.fromCodePoint(0);
  } else {
    return prefix + String.fromCodePoint(lastCodePoint + 1);
  }
};
/**
 * @private
 * Decrement the last character to one character lexigraphically lower. It
 * might not be a valid Unicode character, but that shouldn't matter as we're
 * only using them for range expressions
 * 
 * NOTE: I'm not sure if this is 100% correct as Unicode codepoints can go
 * up to 0x11000, but working on the boundary condition of 0x10FFF isn't
 * a valid codepoint, so JavaScript falls back to code-unit behaviour. For
 * DynamoDB purposes, this may need to be done as a straight Buffer man-
 * ipulation.
 *
 * @param x  the string to increment the last character of
 * @returns the transformed string
 */

const decrementLast = x => {
  var _x$codePointAt2;

  if (x.length === 0) {
    return ''; // This isn't really right but there is nothing lower than the empty string
  }

  const prefix = x.slice(0, x.length - 1);
  const lastCodePoint = (_x$codePointAt2 = x[x.length - 1].codePointAt(0)) !== null && _x$codePointAt2 !== void 0 ? _x$codePointAt2 : 0;

  if (lastCodePoint === 0) {
    return decrementLast(prefix) + String.fromCodePoint(0xFFFF);
  } else {
    return prefix + String.fromCodePoint(lastCodePoint - 1);
  }
};

/**
 * The results of a [[findChildren]] operation.
 */

/**
 * Find all the child objects of a root (top-level) object.
 *
 * The parent collection is determined by the reference in the
 * child collection.
 *
 * This method has a `nextToken`, which is used for pagination - it
 * is returned when there is more values to retrieve. You should
 * write your code to repeatedly call findChildren with the nextToken
 * value of the previous call until it comes back undefined in order
 * to retrieve all the values.
 *
 * @param ctx the context
 * @param childCollectionName name of the child object collection
 * @param rootObjectId the `_id` of the root object
 * @param nextToken the next token from the previous call, or `undefined` if there are no more values
 * @param options the options to control the query
 * @param options.limit number of records to return
 * @param options.scanForward=true true for ascending index order
 * @param options.filter an optional filter expression to apply
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */
async function findChildren(ctx, childCollectionName, rootObjectId, nextToken, options = {}) {
  var _options$scanForward;

  const childCollection = getChildCollection(ctx, childCollectionName);
  const nameMapper = createNameMapper();
  const valueMapper = createValueMapper();
  const {
    parentCollectionName,
    layout: {
      primaryKey: {
        partitionKey,
        sortKey
      }
    }
  } = childCollection;
  const parentId = assemblePrimaryKeyValue(parentCollectionName, rootObjectId, childCollection.layout.indexKeySeparator);
  const partitionKeyExpression = `${nameMapper.map(partitionKey)} = ${valueMapper.map(parentId)}`;
  let sortKeyExpression;

  if (options.range) {
    // for <, <=, >, >=, we can't use the builtin DynamoDB operators because they
    // will include items from other collections naturally. Instead, we construct
    // a barrier for highest and lowest and use the `BETWEEN` operator (similar
    // to findByIdWithChildren)
    switch (options.range.op) {
      case 'gt':
        {
          const childCollectionMin = assemblePrimaryKeyValue(childCollectionName, incrementLast(options.range.value), childCollection.layout.indexKeySeparator);
          const childCollectionMax = assemblePrimaryKeyValue(childCollectionName, '\uFFFF', childCollection.layout.indexKeySeparator);
          sortKeyExpression = `${nameMapper.map(sortKey)} BETWEEN ${valueMapper.map(childCollectionMin)} AND ${valueMapper.map(childCollectionMax)}`;
          break;
        }

      case 'lt':
        {
          const childCollectionMax = assemblePrimaryKeyValue(childCollectionName, decrementLast(options.range.value), childCollection.layout.indexKeySeparator);
          const childCollectionMin = assemblePrimaryKeyValue(childCollectionName, '', childCollection.layout.indexKeySeparator);
          sortKeyExpression = `${nameMapper.map(sortKey)} BETWEEN ${valueMapper.map(childCollectionMin)} AND ${valueMapper.map(childCollectionMax)}`;
          break;
        }

      case 'lte':
        {
          const childCollectionMax = assemblePrimaryKeyValue(childCollectionName, options.range.value, childCollection.layout.indexKeySeparator);
          const childCollectionMin = assemblePrimaryKeyValue(childCollectionName, '', childCollection.layout.indexKeySeparator);
          sortKeyExpression = `${nameMapper.map(sortKey)} BETWEEN ${valueMapper.map(childCollectionMin)} AND ${valueMapper.map(childCollectionMax)}`;
          break;
        }

      case 'gte':
        {
          const childCollectionMin = assemblePrimaryKeyValue(childCollectionName, options.range.value, childCollection.layout.indexKeySeparator);
          const childCollectionMax = assemblePrimaryKeyValue(childCollectionName, '\uFFFF', childCollection.layout.indexKeySeparator);
          sortKeyExpression = `${nameMapper.map(sortKey)} BETWEEN ${valueMapper.map(childCollectionMin)} AND ${valueMapper.map(childCollectionMax)}`;
          break;
        }

      case 'begins_with':
        {
          const childCollectionValue = assemblePrimaryKeyValue(childCollectionName, options.range.value, childCollection.layout.indexKeySeparator);
          sortKeyExpression = `begins_with(${nameMapper.map(sortKey)}, ${valueMapper.map(childCollectionValue)})`;
          break;
        }

      case 'between':
        {
          const childCollectionMin = assemblePrimaryKeyValue(childCollectionName, options.range.min, childCollection.layout.indexKeySeparator);
          const childCollectionMax = assemblePrimaryKeyValue(childCollectionName, options.range.max, childCollection.layout.indexKeySeparator);
          sortKeyExpression = `${nameMapper.map(sortKey)} BETWEEN ${valueMapper.map(childCollectionMin)} AND ${valueMapper.map(childCollectionMax)}`;
          break;
        }

      default:
        throw new InvalidRangeOperatorException('Unknown range operator', options.range.op);
    }
  } else {
    const childCollectionPrefix = assemblePrimaryKeyValue(childCollectionName, '', childCollection.layout.indexKeySeparator);
    sortKeyExpression = `begins_with(${nameMapper.map(sortKey)}, ${valueMapper.map(childCollectionPrefix)})`;
  }

  const keyConditionExpression = `${partitionKeyExpression} AND ${sortKeyExpression}`;
  const request = {
    TableName: childCollection.layout.tableName,
    KeyConditionExpression: keyConditionExpression,
    ExpressionAttributeNames: nameMapper.get(),
    ExpressionAttributeValues: valueMapper.get(),
    ExclusiveStartKey: nextToken,
    Limit: options === null || options === void 0 ? void 0 : options.limit,
    ScanIndexForward: (_options$scanForward = options === null || options === void 0 ? void 0 : options.scanForward) !== null && _options$scanForward !== void 0 ? _options$scanForward : true
  };
  debugDynamo('Query', request);
  const results = await ctx.ddb.query(request).promise();
  return {
    items: (results.Items || []).map(item => unwrap(Converter.unmarshall(item, {
      convertEmptyValues: false
    }))),
    nextToken: results.LastEvaluatedKey
  };
}

/**
 * Retrieve a child item by its `_id` field and its parent `_id`.
 *
 * Child objects can only be directly retrieved with both the parent
 * and child _id. This limitation allows them to be stored without
 * an extra index, and enables the `findChildren` method to retrieve
 * all the children of a root object in the same call.
 *
 * @param context the context object
 * @param collectionName name of the collection to search
 * @param id the `_id` value
 * @param rootObjectId the _id of the root object
 * @returns the stored value, or `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */

async function findChildById(context, collectionName, id, rootObjectId) {
  const collection = getChildCollection(context, collectionName);
  const request = {
    TableName: collection.layout.tableName,
    Key: Converter.marshall({
      [collection.layout.primaryKey.partitionKey]: assemblePrimaryKeyValue(collection.parentCollectionName, rootObjectId, collection.layout.indexKeySeparator),
      [collection.layout.primaryKey.sortKey]: assemblePrimaryKeyValue(collectionName, id, collection.layout.indexKeySeparator)
    }, {
      convertEmptyValues: false
    })
  };
  debugDynamo('GetItem', request);
  const result = await context.ddb.getItem(request).promise();

  if (result.Item) {
    const wrapped = Converter.unmarshall(result.Item, {
      convertEmptyValues: false
    });
    return unwrap(wrapped);
  }

  return undefined;
}

/**
 * Delete a child object using its `_id` field and its parents `_id`.
 *
 * @param context the context object
 * @param collectionName the name of the collection
 * @param id the child object to remove
 * @param rootObjectId the parent object id
 * @param options options to change behaviour of DynamoDB
 * @param options.condition a condition expression blocking the delete operation
 * @returns the deleted object (as stored in the database), or
 * `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */

async function deleteChildById(context, collectionName, id, rootObjectId, options = {}) {
  const collection = getChildCollection(context, collectionName);
  const nameMapper = createNameMapper();
  const valueMapper = createValueMapper();
  let conditionExpression;

  if (options.condition) {
    conditionExpression = parseCompositeCondition(options.condition, {
      nameMapper,
      valueMapper,
      parsePath: []
    });
  }

  const request = {
    TableName: collection.layout.tableName,
    Key: Converter.marshall({
      [collection.layout.primaryKey.partitionKey]: assemblePrimaryKeyValue(collection.parentCollectionName, rootObjectId, collection.layout.indexKeySeparator),
      [collection.layout.primaryKey.sortKey]: assemblePrimaryKeyValue(collectionName, id, collection.layout.indexKeySeparator)
    }, {
      convertEmptyValues: false
    }),
    ReturnValues: 'ALL_OLD',
    ConditionExpression: conditionExpression,
    ExpressionAttributeNames: nameMapper.get(),
    ExpressionAttributeValues: valueMapper.get()
  };
  debugDynamo('DeleteItem', request);
  const result = await context.ddb.deleteItem(request).promise();

  if (result.Attributes) {
    const wrapped = Converter.unmarshall(result.Attributes, {
      convertEmptyValues: false
    });
    return unwrap(wrapped);
  }

  return undefined;
}

/** @internal */

const debug = debug$1('dynaglue:operations:updateById');
/**
 * An update object, where the key paths are specified as keys and the values
 * as the new field values.
 *
 * Keep in mind, that although this is a partial update, you need to specify
 * all the keys for a composite key in an access pattern - it cannot partially
 * update composite index values.
 */

const makeKeyPath = pathOrPathArray => typeof pathOrPathArray === 'string' ? pathOrPathArray.split('.') : pathOrPathArray;
/** @internal */


/**
 * @internal
 * Convert the Updates object to something normalised that
 * is simpler to process internally
 */
const normaliseUpdates = updatesToPerform => {
  if (updatesToPerform.$set || updatesToPerform.$delete) {
    var _changesDocument$$set, _changesDocument$$set2, _changesDocument$$del, _changesDocument$$del2;

    const changesDocument = updatesToPerform;
    return {
      $set: (_changesDocument$$set = (_changesDocument$$set2 = changesDocument.$set) === null || _changesDocument$$set2 === void 0 ? void 0 : _changesDocument$$set2.map(setUpdate => [makeKeyPath(setUpdate[0]), setUpdate[1]])) !== null && _changesDocument$$set !== void 0 ? _changesDocument$$set : [],
      $delete: (_changesDocument$$del = (_changesDocument$$del2 = changesDocument.$delete) === null || _changesDocument$$del2 === void 0 ? void 0 : _changesDocument$$del2.map(makeKeyPath)) !== null && _changesDocument$$del !== void 0 ? _changesDocument$$del : []
    };
  } else {
    const updatesDocument = updatesToPerform;
    return {
      $set: Object.entries(updatesDocument).map(([key, value]) => [makeKeyPath(key), value]),
      $delete: []
    };
  }
};
/**
 * @internal
 */

const extractUpdateKeyPaths = changes => [...changes.$set.map(([path]) => path), ...changes.$delete];
/**
 * @internal
 */

const getValueForUpdatePath = (matchingUpdatePath, keyPath, changes) => {
  // Work out if this was a $set update (in which case we get the index) or
  // a delete path (which we perform by deduction)
  const setPathIndex = changes.$set.findIndex(([setPath]) => isEqualKey(setPath, matchingUpdatePath));
  let value = setPathIndex >= 0 ? changes.$set[setPathIndex][1] : undefined;

  if (setPathIndex >= 0 && keyPath.length !== matchingUpdatePath.length) {
    const difference = keyPath.slice(matchingUpdatePath.length);
    value = get(value, difference);
  }

  return value;
};
/**
 * @internal
 */

const createUpdateActionForKey = (collectionName, keyType, keyPaths, indexLayout, changes, separator) => {
  const updateKeyPaths = extractUpdateKeyPaths(changes);
  const matchingUpdatePaths = keyPaths.map(partitionKey => findMatchingPath(updateKeyPaths, partitionKey));
  const attributeName = keyType === 'sort' ? indexLayout.sortKey : indexLayout.partitionKey;
  debug('createUpdateActionForKey: collection=%s keyType=%s keyPaths=%o attributeName=%s', collectionName, keyType, keyPaths, attributeName);

  if (matchingUpdatePaths.every(updatePath => typeof updatePath === 'undefined')) {
    debug('createUpdateActionForKey: no updates to %s key in collection %s', keyType, collectionName);
    return undefined;
  }

  debug('createUpdateActionForKey: key to be updated matchingUpdatePaths=%o', matchingUpdatePaths);
  const updateValues = keyPaths.map((keyPath, index) => {
    const matchingUpdatePath = matchingUpdatePaths[index];

    if (!matchingUpdatePath) {
      return undefined;
    }

    return getValueForUpdatePath(matchingUpdatePath, keyPath, changes);
  });
  return {
    attributeName,
    value: assembleIndexedValue(keyType, collectionName, updateValues, separator),
    valueErasure: updateValues.every(value => typeof value === 'undefined')
  };
};
/**
 * @internal
 *
 * Create the update action for a TTL key path
 */

const createUpdateActionForTTLKey = (attributeName, keyPath, updates) => {
  const updateKeyPaths = extractUpdateKeyPaths(updates);
  const matchingUpdatePath = findMatchingPath(updateKeyPaths, keyPath);

  if (matchingUpdatePath) {
    const value = getValueForUpdatePath(matchingUpdatePath, keyPath, updates);
    return {
      attributeName,
      value: value ? transformTTLValue(value) : undefined
    };
  }

  return undefined;
};
/**
 * @internal
 */

const findCollectionIndex = (collection, indexName) => {
  var _collection$layout$fi;

  const layout = (_collection$layout$fi = collection.layout.findKeys) === null || _collection$layout$fi === void 0 ? void 0 : _collection$layout$fi.find(fk => fk.indexName === indexName);

  if (!layout) {
    throw new IndexNotFoundException(indexName);
  }

  return layout;
};
/**
 * @internal
 */

/**
 * @internal
 *
 * Given the set of updates, create the SET and DELETE
 * actions for the access patterns that also have to be
 * changed.
 */
const mapAccessPatterns = (collection, {
  nameMapper,
  valueMapper
}, changes) => {
  const expressionSetActions = [];
  const expressionDeleteActions = [];
  const {
    accessPatterns = [],
    ttlKeyPath
  } = collection;

  for (const {
    indexName,
    partitionKeys,
    sortKeys
  } of accessPatterns) {
    let partitionKeyUpdateSet = undefined;
    let sortKeyUpdated = false;
    const layout = findCollectionIndex(collection, indexName);

    if (partitionKeys.length > 0) {
      const update = createUpdateActionForKey(collection.name, 'partition', partitionKeys, layout, changes, collection.layout.indexKeySeparator);

      if (update) {
        partitionKeyUpdateSet = !update.valueErasure;
        debug('mapAccessPatterns: adding set/delete action for partition key in collection %s: %o', collection.name, update);
        const nameMapping = nameMapper.map(update.attributeName);
        const valueMapping = valueMapper.map(update.value);
        expressionSetActions.push(`${nameMapping} = ${valueMapping}`);
      }
    }

    if (sortKeys && sortKeys.length > 0) {
      const update = createUpdateActionForKey(collection.name, 'sort', sortKeys, layout, changes, collection.layout.indexKeySeparator);

      if (update) {
        debug('mapAccessPatterns: adding set/delete action for sort key in collection %s: %o', collection.name, update);
        sortKeyUpdated = true;

        if (typeof update.value !== 'undefined') {
          const nameMapping = nameMapper.map(update.attributeName);
          const valueMapping = valueMapper.map(update.value);
          expressionSetActions.push(`${nameMapping} = ${valueMapping}`);
        } else {
          const nameMapping = nameMapper.map(update.attributeName);
          expressionDeleteActions.push(nameMapping);
        }
      }
    } else if (typeof partitionKeyUpdateSet !== 'undefined' && layout.sortKey) {
      // When only primary key has indexed paths (i.e. partitionKeys.length > 0,
      // sortKeys.length === 0, and we applied an update for the partitionKey)
      // make an empty update to the sort key to in-case it wasn't populated
      const nameMapping = nameMapper.map(layout.sortKey);

      if (partitionKeyUpdateSet) {
        const valueMapping = valueMapper.map(assembleIndexedValue('sort', collection.name, [], collection.layout.indexKeySeparator));
        expressionSetActions.push(`${nameMapping} = ${valueMapping}`);
      } else {
        expressionDeleteActions.push(nameMapping);
      }
    } // In the case the sort key is updated but there is no indexed partition key
    // paths, make sure the partition key gets a value written


    if (sortKeyUpdated && (!partitionKeys || partitionKeys.length === 0)) {
      const nameMapping = nameMapper.map(layout.partitionKey);
      const valueMapping = valueMapper.map(assembleIndexedValue('partition', collection.name, [], collection.layout.indexKeySeparator));
      expressionSetActions.push(`${nameMapping} = ${valueMapping}`);
    }
  }

  if (ttlKeyPath) {
    const updateAction = createUpdateActionForTTLKey( // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    collection.layout.ttlAttribute, // we've already asserted this in context creation
    ttlKeyPath, changes);

    if (updateAction) {
      const nameMapping = nameMapper.map(updateAction.attributeName);

      if (updateAction.value) {
        const valueMapping = valueMapper.map(updateAction.value);
        expressionSetActions.push(`${nameMapping} = ${valueMapping}`);
      } else {
        expressionDeleteActions.push(nameMapping);
      }
    }
  }

  return {
    setActions: expressionSetActions,
    deleteActions: expressionDeleteActions
  };
};
/**
 * @internal
 *
 * Performs an update operation for the given collection and key
 * value. Shares most of the code between updateById and updateChildById
 *
 */

async function updateInternal(ctx, collection, key, updatesToPerform, options) {
  const changes = normaliseUpdates(updatesToPerform);

  if (changes.$set.length === 0 && changes.$delete.length === 0) {
    throw new InvalidUpdatesException('There must be at least one update path in the updates object');
  }

  const nameMapper = createNameMapper();
  const valueMapper = createValueMapper();
  let expressionSetActions = [];
  let expressionDeleteActions = [];

  for (const [path, newValue] of changes.$set.values()) {
    if (typeof newValue === 'undefined') {
      throw new InvalidUpdateValueException(path.join('.'), 'value must not be undefined');
    }

    const valueName = valueMapper.map(newValue);
    const expressionAttributeNameParts = [nameMapper.map('value', '#value'), ...path.map(part => nameMapper.map(part))];
    expressionSetActions.push(`${expressionAttributeNameParts.join('.')} = ${valueName}`);
  }

  for (const path of changes.$delete) {
    const expressionAttributeNameParts = [nameMapper.map('value', '#value'), ...path.map(part => nameMapper.map(part))];
    expressionDeleteActions.push(`${expressionAttributeNameParts.join('.')}`);
  }

  const {
    setActions: additionalSetActions,
    deleteActions: additionalDeleteActions
  } = mapAccessPatterns(collection, {
    nameMapper,
    valueMapper
  }, changes);
  expressionSetActions = [...expressionSetActions, ...additionalSetActions];
  expressionDeleteActions = [...expressionDeleteActions, ...additionalDeleteActions];
  let conditionExpression;

  if (options.condition) {
    conditionExpression = parseCompositeCondition(options.condition, {
      nameMapper,
      valueMapper,
      parsePath: []
    });
  }

  const expressionAttributeNames = nameMapper.get();
  const expressionAttributeValues = valueMapper.get();
  const updateExpression = (expressionSetActions.length ? ` SET ${expressionSetActions.join(', ')}` : '') + (expressionDeleteActions.length ? ` REMOVE ${expressionDeleteActions.join(', ')}` : '');
  const updateItem = {
    TableName: collection.layout.tableName,
    Key: key,
    ReturnValues: 'ALL_NEW',
    ExpressionAttributeNames: expressionAttributeNames,
    ExpressionAttributeValues: expressionAttributeValues,
    UpdateExpression: updateExpression.trim(),
    ConditionExpression: conditionExpression
  };
  debugDynamo('UpdateItem', updateItem);
  const result = await ctx.ddb.updateItem(updateItem).promise();
  const unmarshalledAttributes = Converter.unmarshall(result.Attributes, {
    convertEmptyValues: false
  });
  const updatedDocument = unwrap(unmarshalledAttributes);
  return updatedDocument;
}
/**
 * Update a document using its `_id`.
 *
 * This operation allows you to do a partial update of a collection document i.e. without
 * specifying all the values (it uses DynamoDB`s `UpdateItem` operation).
 *
 * At this time, the `updates` value just updates specified key paths on the target document.
 *
 * If some of the update key paths are indexed values, the indexes will also be updated. Because
 * of this, you must specify all the key values in an access pattern to ensure indexes are
 * updated consistently.
 *
 * @param ctx the context
 * @param collectionName the collection to update
 * @param objectId the `_id` value of the object to update
 * @param updates the set of updates to apply.
 * @returns the updated object value in its entirety.
 * @throws {CollectionNotFoundException} collection not found
 * @throws {InvalidUpdatesException} thrown when the updates object is invalid or incomplete
 * @throws {InvalidUpdateValueException} thrown when one of the update values is an invalid type
 */

async function updateById(ctx, collectionName, objectId, updates, options = {}) {
  const collection = getRootCollection(ctx, collectionName);
  const key = {
    [collection.layout.primaryKey.partitionKey]: {
      S: assemblePrimaryKeyValue(collectionName, objectId, collection.layout.indexKeySeparator)
    },
    [collection.layout.primaryKey.sortKey]: {
      S: assemblePrimaryKeyValue(collectionName, objectId, collection.layout.indexKeySeparator)
    }
  };
  return updateInternal(ctx, collection, key, updates, options);
}

/**
 * Update a child document using its `_id` and parent `_id`.
 *
 * This operation allows you to do a partial update of a collection document i.e. without
 * specifying all the values (it uses DynamoDB`s `UpdateItem` operation).
 *
 * At this time, the `updates` value just updates specified key paths on the target document.
 *
 * If some of the update key paths are indexed values, the indexes will also be updated. Because
 * of this, you must specify all the key values in an access pattern to ensure indexes are
 * updated consistently.
 *
 * @param ctx the context
 * @param collectionName the collection to update
 * @param objectId the `_id` value of the object to update
 * @param parentObjectId: the `_id` value of the parent object
 * @param updates the set of updates to apply.
 * @param options options to apply
 * @param options.condition an optional conditional expression that must be satifisfied for the update to proceed
 * @returns the updated object value in its entirety.
 * @throws {CollectionNotFoundException} collection not found
 * @throws {InvalidUpdatesException} thrown when the updates object is invalid or incomplete
 * @throws {InvalidUpdateValueException} thrown when one of the update values is an invalid type
 */
async function updateChildById(context, collectionName, objectId, parentObjectId, updates, options = {}) {
  const collection = getChildCollection(context, collectionName);
  const key = {
    [collection.layout.primaryKey.partitionKey]: {
      S: assemblePrimaryKeyValue(collection.parentCollectionName, parentObjectId, collection.layout.indexKeySeparator)
    },
    [collection.layout.primaryKey.sortKey]: {
      S: assemblePrimaryKeyValue(collectionName, objectId, collection.layout.indexKeySeparator)
    }
  };
  return updateInternal(context, collection, key, updates, options);
}

/**
 * Find a root object with one or more of its child collections.

 * It is used to implement the **adjacency list** access approach,
 * which is used to reduce round-trips by loading a parent
 * object and its associated child objects in the same call. 
 *
 * (Because of this, it's important to understand adjacency lists
 * and why you would have modelled your collections in a
 * parent-child relationship to begin with in order to get
 * the desired performance outcomes)
 *
 * You can restrict the child collections to return by specifying
 * the `childCollectionNames` parameter. Leaving it blank retrieves
 * all the child objects.
 *
 * ## Performance
 *
 * Adjacency lists work by storing all the related items in the
 * same partition (key) i.e. with the same object ID as the 
 * root object, and then keying the individual items in the
 * sort key. 
 *
 * We can then use a `Query` call to return the root object
 * and child objects together. This means that the performance 
 * and behaviour depends on however the items are stored in the index. 
 *
 * For this reason it's vital to have
 * some understanding of how it is implemented, as well as
 * be aware of your most important access patterns, in order
 * to get the best performance (especially when you have
 * many child items or multiple child collections).
 *
 * ## How it works
 *
 * This method works by using a DynamoDB `Query` on the root
 * object ID (in the partition key). It restricts the child
 * objects returned by using a `BETWEEN` filter on the sort
 * key, bounded by lexigraphically lowest and highest collection
 * names. 
 *
 * To ensure that only the specified collections are returned,
 * this is then filtered with a filter expression on the desired
 * child collection names. This is done with a `FilterExpression`
 * in the `Query` call.
 *
 * In the ideal scenario, all the 
 * child objects are small enough be returned within DynamoDB's query limit.
 * In this case the `root` value will contain the specified parent object, and 
 * the `children` value will be a key-value index on the child collection name
 * to its associated child objects.
 *
 * However, if the number of items exceeds the DynamoDB query limit (1MB
 * of total items scanned in the index), you will need to use
 * the `nextToken` to retrieve the next set of results. Because of
 * this, the `root` object may not be returned on the first call,
 * so you should check for it on each set of returned results.
 *
 * ## Working through an example
 *
 * Say you have a root object collection called *users*,
 * with two child collections, `addresses` and `credentials`.
 *
 * We have two main access patterns:
 * 1. Retrieve a user with all their addresses
 * 2. Retrieve a user will all their credentials
 *
 * ### Retrieving all the addresses
 *
 * For 1., where you request with child collection `addresses` only, a query like
 * this will be given to DynamoDB:
 *
 * ```
 * KeyConditionExpression: pk = "users|-|<userId>" AND sk BETWEEN "addresses|-|" AND 'users|-|'
 * FilterExpression: type IN ("users", "addresses")
 * ```
 *
 * Because of the way DynamoDB queries work, it will internally
 * scan over all the child objects (including the *credentials* objects).
 * The unwanted `credentials` objects will be filtered out, but
 * they will still count internally to the DynamoDB Query page limit
 * (about 1MB).
 *
 * The addresses objects will be returned first (the default ordering
 * goes in lexigraphic order), unless you specify `scanIndexForward: false`
 * to start at the end (where the *users* object is).
 *
 * If there isn't many unwanted *credentials*, the performance impact
 * will be neglible. 
 *
 * If there any many *credentials* objects however, there could be many
 * blank pages returned until the root object is returned. 
 *
 * ### Addressing the performance
 *
 * Firstly, we should change the scan behaviour to scan the index backwards.
 * This way, we'll get the root object in the first page. But DynamoDB
 * will still be scanning over the credentials objects internally.
 *
 * In cases like these, it may be better to break your request down into two
 * calls:
 *
 * * `findById` to return the root *users* object
 * * `findChildren` to return all the *addresses* object.
 *
 * However, this would negate the point of using a adjacency list
 * in the first place.
 *
 * If we look at our access patterns, we never want to retrieve 
 * the credentials and addresses objects at the same time, so we
 * really want the *users* root object to appear between the two
 * sets of child objects.
 *
 * We could achieve this by changing the collection names so
 * that it is between the other two by adding a prefix to each collection name:
 *
 * ```
 * 0addresses
 * 1users
 * 2credentials
 * ```
 *
 * Then, we can fulfil (1) by requesting *1addresses* and scanning backwards
 * from *users*. Similarly, we can fulfill (2) by requesting *credentials*
 * and scanning in the default index order. In both cases, we won't be
 * scanning over unnecessary objects.
 *
 * @param ctx the dynaglue context object
 * @param rootObjectCollectionName the root object collection
 * @param rootObjectId the root object ID
 * @param childCollectionNames the child collections to return - defaults to all child collections
 * @param nextToken the pagination token, to request the next page of results
 * @param options options to control the limit, scan order and extra filtering
 * @returns the root and/or child objects found, along with a `nextToken` if there is more results
 *
 */
const findByIdWithChildren = async (ctx, rootCollectionName, rootObjectId, childCollectionNames, nextToken, options = {}) => {
  var _options$scanForward;

  const collection = getCollection(ctx, rootCollectionName);
  let childCollections;

  if (childCollectionNames) {
    childCollections = childCollectionNames.map(name => getChildCollection(ctx, name));

    for (const childCollection of childCollections) {
      if (childCollection.parentCollectionName !== rootCollectionName) {
        throw new CollectionNotFoundException(childCollection.name);
      }
    }
  } else {
    childCollections = [...ctx.childDefinitions.values()].filter(collection => collection.parentCollectionName === rootCollectionName);
  }

  const allChildCollectionNames = childCollections.map(collection => collection.name);
  const allCollectionNames = [rootCollectionName, ...allChildCollectionNames].sort();
  const firstCollection = allCollectionNames[0];
  const lastCollection = allCollectionNames[allCollectionNames.length - 1];
  const lastCollectionBound = assemblePrimaryKeyValue(lastCollection, '\uFFFF', collection.layout.indexKeySeparator);
  const nameMapper = createNameMapper();
  const valueMapper = createValueMapper();
  const {
    partitionKey,
    sortKey
  } = collection.layout.primaryKey;
  const keyConditionExpression = `${nameMapper.map(partitionKey)} = ${valueMapper.map(assemblePrimaryKeyValue(rootCollectionName, rootObjectId, collection.layout.indexKeySeparator))} ` + `AND ${nameMapper.map(sortKey)} BETWEEN ${valueMapper.map(assemblePrimaryKeyValue(firstCollection, '', collection.layout.indexKeySeparator))} AND ${valueMapper.map(lastCollectionBound)}`;
  const filterExpression = `${nameMapper.map('type')} IN (${allCollectionNames.map(c => valueMapper.map(c)).join(',')})`;
  const request = {
    TableName: collection.layout.tableName,
    KeyConditionExpression: keyConditionExpression,
    ExpressionAttributeNames: nameMapper.get(),
    ExpressionAttributeValues: valueMapper.get(),
    ExclusiveStartKey: nextToken,
    Limit: options === null || options === void 0 ? void 0 : options.limit,
    ScanIndexForward: (_options$scanForward = options === null || options === void 0 ? void 0 : options.scanForward) !== null && _options$scanForward !== void 0 ? _options$scanForward : true,
    FilterExpression: filterExpression
  };
  debugDynamo('Query', request);
  const {
    Items = [],
    LastEvaluatedKey
  } = await ctx.ddb.query(request).promise();
  let root;
  const children = fromPairs(allChildCollectionNames.map(name => [name, []]));

  for (const item of Items) {
    const attributes = Converter.unmarshall(item, {
      convertEmptyValues: false
    });

    if (attributes.type === rootCollectionName) {
      root = unwrap(attributes);
    } else {
      var _children$attributes$;

      children[attributes.type] = (_children$attributes$ = children[attributes.type]) !== null && _children$attributes$ !== void 0 ? _children$attributes$ : [];
      children[attributes.type].push(unwrap(attributes));
    }
  }

  return {
    root,
    children,
    nextToken: LastEvaluatedKey
  };
};

/**
 * @internal
 *
 * The ID of a root or child item.
 *
 */

/**
 * @internal
 *
 * Parse a dynaglue key. For use with BatchGetItem/BatchWriteItem interactions,
 * which return unprocessed items in a random order, requiring us to disassemble
 * our request in order to work out what to resubmit.
 *
 */
const parseKey = (layout, key) => {
  var _layout$indexKeySepar, _layout$indexKeySepar2;

  const partitionKey = Converter.output(key[layout.primaryKey.partitionKey], {
    convertEmptyValues: false
  });
  const sortKey = Converter.output(key[layout.primaryKey.sortKey], {
    convertEmptyValues: false
  }); // The following checks are for sanity and should not occur in any real application that
  // has setup the layout correctly

  if (!partitionKey || !sortKey) throw new InternalProcessingException(`Selected layout could not find key names ${JSON.stringify(key)} for table ${layout.tableName}`);
  if (typeof partitionKey !== 'string') throw new InternalProcessingException(`Partition key ${layout.primaryKey.partitionKey} for table ${layout.tableName} is not a string`);
  if (typeof sortKey !== 'string') throw new InternalProcessingException(`Sort key ${layout.primaryKey.sortKey} for table ${layout.tableName} is not a string`);
  const [, parentId] = partitionKey.split((_layout$indexKeySepar = layout.indexKeySeparator) !== null && _layout$indexKeySepar !== void 0 ? _layout$indexKeySepar : SEPARATOR, 2);
  const [childCollection, id] = sortKey.split((_layout$indexKeySepar2 = layout.indexKeySeparator) !== null && _layout$indexKeySepar2 !== void 0 ? _layout$indexKeySepar2 : SEPARATOR, 2);
  return {
    collection: childCollection,
    parentId: parentId !== id ? parentId : undefined,
    id: id
  };
};

/**
 * The collection and ID of a root or child
 * item to retrieve with #transactFindByIds
 */

const batchFindByIds = async (ctx, items, options) => {
  if (items.length === 0) {
    throw new InvalidFindDescriptorException('At least one find descriptor must be specified');
  } else if (items.length > 100) {
    throw new InvalidFindDescriptorException('No more than 100 find descriptors can be specified to batchFindByIds');
  } // Map tables to their 'predicted' layouts - this is needed
  // because of the abject uselessness of the UnprocessedKeys structure, which
  // returns unprocessed items in random order arranged only
  // by table name.
  //
  // Because we could be accessing multiple tables, we need
  // the layout to discern the key names when parsing the
  // UnprocessedKeys map.


  const tableLayoutMapping = new Map();
  const tableRequestItemTuples = items.map(({
    collection,
    id,
    parentId
  }) => {
    const collectionDefinition = parentId ? getChildCollection(ctx, collection) : getRootCollection(ctx, collection);
    tableLayoutMapping.set(collectionDefinition.layout.tableName, collectionDefinition.layout);
    const {
      layout: {
        tableName,
        primaryKey,
        indexKeySeparator
      }
    } = collectionDefinition;
    return [tableName, {
      [primaryKey.partitionKey]: Converter.input(assemblePrimaryKeyValue(collectionDefinition.type === 'child' ? collectionDefinition.parentCollectionName : collectionDefinition.name, parentId ? parentId : id, indexKeySeparator), {
        convertEmptyValues: false
      }),
      [primaryKey.sortKey]: Converter.input(assemblePrimaryKeyValue(collection, id, indexKeySeparator), {
        convertEmptyValues: false
      })
    }];
  });
  const requestItems = tableRequestItemTuples.reduce((req, tuple) => {
    var _req$tuple$, _options$consistentRe;

    const keyAndsAttrs = (_req$tuple$ = req[tuple[0]]) !== null && _req$tuple$ !== void 0 ? _req$tuple$ : {
      ConsistentRead: ((_options$consistentRe = options === null || options === void 0 ? void 0 : options.consistentReadTableNames) !== null && _options$consistentRe !== void 0 ? _options$consistentRe : []).includes(tuple[0]),
      Keys: []
    };
    req[tuple[0]] = keyAndsAttrs;
    keyAndsAttrs.Keys.push(tuple[1]);
    return req;
  }, {});
  const request = {
    RequestItems: requestItems
  };
  debugDynamo('batchGetItem', request);
  const {
    Responses = {},
    UnprocessedKeys = {}
  } = await ctx.ddb.batchGetItem(request).promise();
  const response = {
    documentsByCollection: {},
    unprocessedDescriptors: []
  };

  for (const items of Object.values(Responses)) {
    for (const item of items) {
      var _response$documentsBy;

      const unmarshalled = Converter.unmarshall(item, {
        convertEmptyValues: false
      });
      const collection = getCollection(ctx, unmarshalled.type);
      const document = unwrap(unmarshalled);
      const collectionMap = (_response$documentsBy = response.documentsByCollection[collection.name]) !== null && _response$documentsBy !== void 0 ? _response$documentsBy : [];
      response.documentsByCollection[collection.name] = collectionMap;
      collectionMap.push(document);
    }
  }

  for (const [tableName, unprocessed] of Object.entries(UnprocessedKeys)) {
    const tableMapping = tableLayoutMapping.get(tableName);

    if (!tableMapping) {
      throw new InternalProcessingException(`Could not find table mapping for ${tableName} while parsing UnprocessedKeys`);
    }

    for (const key of unprocessed.Keys) {
      const descriptor = parseKey(tableMapping, key);
      response.unprocessedDescriptors.push(descriptor);
    }
  }

  return response;
};

/**
 * A replace (PutItem) request to perform with
 * [[batchReplaceDelete]].
 *
 * Performs the equivalent operation as [[replace]],
 * which is a full insert or update on the primary
 * key of the document.
 */

const batchReplaceDelete = async (ctx, descriptors) => {
  if (descriptors.length === 0) {
    throw new InvalidBatchReplaceDeleteDescriptorException('You must specify at least one replace or delete descriptor');
  } // Because we could be accessing multiple tables, we need
  // the layout to discern the key names when parsing the
  // UnprocessedKeys map.


  const tableLayoutMapping = new Map();
  const tableRequestItemTuples = descriptors.map(descriptor => {
    const collection = getCollection(ctx, descriptor.collection);
    tableLayoutMapping.set(collection.layout.tableName, collection.layout);
    let request;

    if (descriptor.op === 'replace') {
      // replace item
      request = {
        PutRequest: {
          Item: Converter.marshall(toWrapped(collection, descriptor.replaceItem), {
            convertEmptyValues: false
          })
        }
      };
    } else {
      var _descriptor$parentId;

      if (collection.type === 'child' && !descriptor.parentId) {
        throw new InvalidBatchReplaceDeleteDescriptorException('BatchDeleteDescriptor must specify parentId for child collections', {
          collectionName: descriptor.collection,
          id: descriptor.id
        });
      } else if (collection.type !== 'child' && descriptor.parentId) {
        throw new InvalidBatchReplaceDeleteDescriptorException('BatchDeleteDescriptor must not specify parentId for root collections', {
          collectionname: descriptor.collection,
          id: descriptor.id
        });
      }

      const partitionKeyValue = assemblePrimaryKeyValue(collection.type === 'child' ? collection.parentCollectionName : collection.name, (_descriptor$parentId = descriptor.parentId) !== null && _descriptor$parentId !== void 0 ? _descriptor$parentId : descriptor.id, collection.layout.indexKeySeparator);
      const sortKeyValue = assemblePrimaryKeyValue(collection.name, descriptor.id, collection.layout.indexKeySeparator);
      request = {
        DeleteRequest: {
          Key: Converter.marshall({
            [collection.layout.primaryKey.partitionKey]: partitionKeyValue,
            [collection.layout.primaryKey.sortKey]: sortKeyValue
          }, {
            convertEmptyValues: false
          })
        }
      };
    }

    return [collection.layout.tableName, request];
  });
  const requestItems = tableRequestItemTuples.reduce((riMap, [table, request]) => {
    var _riMap$table;

    const items = (_riMap$table = riMap[table]) !== null && _riMap$table !== void 0 ? _riMap$table : [];
    riMap[table] = items;
    items.push(request);
    return riMap;
  }, {});
  const request = {
    RequestItems: requestItems
  };
  debugDynamo('batchWriteItem', request);
  const {
    UnprocessedItems = {}
  } = await ctx.ddb.batchWriteItem(request).promise();
  const unprocessedDescriptors = [];

  for (const [tableName, unprocessed] of Object.entries(UnprocessedItems)) {
    const tableMapping = tableLayoutMapping.get(tableName);

    if (!tableMapping) {
      throw new InternalProcessingException(`Could not find table mapping for ${tableName} while parsing UnprocessedKeys`);
    }

    for (const item of unprocessed) {
      const {
        PutRequest,
        DeleteRequest
      } = item;

      if (PutRequest) {
        const key = parseKey(tableMapping, PutRequest.Item);
        const wrapped = Converter.unmarshall(PutRequest.Item, {
          convertEmptyValues: false
        });
        const document = unwrap(wrapped);
        unprocessedDescriptors.push({
          op: 'replace',
          replaceItem: document,
          collection: key.collection
        });
      } else if (DeleteRequest) {
        const key = parseKey(tableMapping, DeleteRequest.Key);
        unprocessedDescriptors.push({
          op: 'delete',
          ...key
        });
      }
    }
  }

  return {
    unprocessedDescriptors
  };
};

export { CollectionNotFoundException, ConfigurationException, ConflictException, IndexNotFoundException, InvalidBatchReplaceDeleteDescriptorException, InvalidCompositeConditionException, InvalidIdException, InvalidIndexedFieldValueException, InvalidParentIdException, InvalidQueryException, InvalidUpdateValueException, InvalidUpdatesException, batchFindByIds, batchReplaceDelete, createContext, deleteById, deleteChildById, find, findById, findByIdWithChildren, findChildById, findChildren, insert, replace, updateById, updateChildById };
